from pyetp.resqml_objects.generated import (
    APIGammaRayMeasure,
    APIGammaRayUom,
    APIGravityMeasure,
    APIGravityUom,
    APINeutronMeasure,
    APINeutronUom,
    AbsorbedDoseMeasure,
    AbsorbedDoseUom,
    AbstractActivityParameter,
    AbstractBooleanArray,
    AbstractCRSType,
    AbstractCitedDataObject,
    AbstractColumnLayerGridGeometry,
    AbstractColumnLayerGridRepresentation,
    AbstractContactInterpretationPart,
    AbstractContactRepresentationPart,
    AbstractContextualObject,
    AbstractCoordinateOperationType,
    AbstractCoordinateSystemType,
    AbstractDQ_Element_Type,
    AbstractDQ_PositionalAccuracy_Type,
    AbstractDQ_Result_Type,
    AbstractDataObject,
    AbstractDatumType,
    AbstractDoubleArray,
    AbstractEX_GeographicExtent_Type,
    AbstractFeature,
    AbstractFeatureInterpretation,
    AbstractGMLType,
    AbstractGeneralConversionType,
    AbstractGeneralDerivedCRSType,
    AbstractGeologicFeature,
    AbstractGeometry,
    AbstractGridGeometry,
    AbstractGridRepresentation,
    AbstractIntegerArray,
    AbstractLocal3dCrs,
    AbstractObject,
    AbstractObject_Type,
    AbstractOrganizationInterpretation,
    AbstractParameterKey,
    AbstractParametricLineArray,
    AbstractParametricLineGeometry,
    AbstractParentWindow,
    AbstractPlaneGeometry,
    AbstractPoint3dArray,
    AbstractProjectedCrs,
    AbstractProperty,
    AbstractPropertyKind,
    AbstractPropertyLookup,
    AbstractRepresentation,
    AbstractResqmlDataObject,
    AbstractSeismicCoordinates,
    AbstractSeismicSurveyFeature,
    AbstractStratigraphicOrganizationInterpretation,
    AbstractSurfaceFrameworkRepresentation,
    AbstractSurfaceRepresentation,
    AbstractTechnicalFeature,
    AbstractTimeObjectType,
    AbstractTimePrimitiveType,
    AbstractTruncatedColumnLayerGridRepresentation,
    AbstractValueArray,
    AbstractValuesProperty,
    AbstractVerticalCrs,
    Activation,
    Activity,
    ActivityOfRadioactivityMeasure,
    ActivityOfRadioactivityUom,
    ActivityTemplate,
    AdditionalGridPoints,
    AdditionalGridTopology,
    AggregationType,
    AmountOfSubstanceMeasure,
    AmountOfSubstancePerAmountOfSubstanceMeasure,
    AmountOfSubstancePerAmountOfSubstanceUom,
    AmountOfSubstancePerAreaMeasure,
    AmountOfSubstancePerAreaUom,
    AmountOfSubstancePerTimeMeasure,
    AmountOfSubstancePerTimePerAreaMeasure,
    AmountOfSubstancePerTimePerAreaUom,
    AmountOfSubstancePerTimeUom,
    AmountOfSubstancePerVolumeMeasure,
    AmountOfSubstancePerVolumeUom,
    AmountOfSubstanceUom,
    AnglePerLengthMeasure,
    AnglePerLengthUom,
    AnglePerVolumeMeasure,
    AnglePerVolumeUom,
    AngularAccelerationMeasure,
    AngularAccelerationUom,
    AngularVelocityMeasure,
    AngularVelocityUom,
    AreaMeasure,
    AreaPerAmountOfSubstanceMeasure,
    AreaPerAmountOfSubstanceUom,
    AreaPerAreaMeasure,
    AreaPerAreaUom,
    AreaPerMassMeasure,
    AreaPerMassUom,
    AreaPerTimeMeasure,
    AreaPerTimeUom,
    AreaPerVolumeMeasure,
    AreaPerVolumeUom,
    AreaUom,
    AttenuationPerFrequencyIntervalMeasure,
    AttenuationPerFrequencyIntervalUom,
    AxisOrder2d,
    BinaryContactInterpretationPart,
    BlockedWellboreRepresentation,
    Boolean,
    BooleanArrayFromDiscretePropertyArray,
    BooleanArrayFromIndexArray,
    BooleanConstantArray,
    BooleanHdf5Array,
    Boolean_PropertyType,
    BoundaryFeature,
    BoundaryFeatureInterpretation,
    BoundaryRelation,
    CI_Address,
    CI_Address_PropertyType,
    CI_Address_Type,
    CI_Citation,
    CI_Citation_PropertyType,
    CI_Citation_Type,
    CI_Contact,
    CI_Contact_PropertyType,
    CI_Contact_Type,
    CI_Date,
    CI_DateTypeCode,
    CI_DateTypeCode_PropertyType,
    CI_Date_PropertyType,
    CI_Date_Type,
    CI_OnLineFunctionCode,
    CI_OnLineFunctionCode_PropertyType,
    CI_OnlineResource,
    CI_OnlineResource_PropertyType,
    CI_OnlineResource_Type,
    CI_PresentationFormCode,
    CI_PresentationFormCode_PropertyType,
    CI_ResponsibleParty,
    CI_ResponsibleParty_PropertyType,
    CI_ResponsibleParty_Type,
    CI_RoleCode,
    CI_RoleCode_PropertyType,
    CI_Series,
    CI_Series_PropertyType,
    CI_Series_Type,
    CI_Telephone,
    CI_Telephone_PropertyType,
    CI_Telephone_Type,
    CRSPropertyType,
    CapacitanceMeasure,
    CapacitanceUom,
    CartesianCSPropertyType,
    CartesianCSType,
    CartesianCS_1,
    CategoricalProperty,
    CategoricalPropertySeries,
    CellFluidPhaseUnits,
    CellOverlap,
    CellParentWindow,
    CellShape,
    CellStratigraphicUnits,
    CharacterString,
    CharacterString_PropertyType,
    ChronostratigraphicRank,
    Citation,
    CodeListValue_Type,
    CodeType,
    CodeWithAuthorityType,
    ColumnLayerParentWindow,
    ColumnLayerSplitColumnEdges,
    ColumnLayerSplitCoordinateLines,
    ColumnLayerSubnodeTopology,
    ColumnShape,
    ColumnSubnodePatch,
    CommentProperty,
    CommentPropertySeries,
    ConnectionInterpretations,
    ContactElementReference,
    ContactIdentity,
    ContactMode,
    ContactPatch,
    ContactRelationship,
    ContactRepresentationReference,
    ContactSide,
    ContactVerb,
    ContinuousProperty,
    ContinuousPropertySeries,
    CoordinateSystemAxis,
    CoordinateSystemAxisPropertyType,
    CoordinateSystemAxisType,
    CustomData,
    DQ_EvaluationMethodTypeCode,
    DQ_EvaluationMethodTypeCode_PropertyType,
    DQ_Result_PropertyType,
    DataObjectParameter,
    DataObjectReference,
    DataTransferSpeedMeasure,
    DataTransferSpeedUom,
    Date,
    DateTime,
    DateTime_PropertyType,
    Date_PropertyType,
    Definition,
    DefinitionBaseType,
    DefinitionType,
    DepositionMode,
    DeviationSurveyRepresentation,
    DiffusionCoefficientMeasure,
    DiffusionCoefficientUom,
    DigitalStorageMeasure,
    DigitalStorageUom,
    DimensionlessMeasure,
    DimensionlessUom,
    DipoleMomentMeasure,
    DipoleMomentUom,
    DiscreteProperty,
    DiscretePropertySeries,
    Domain,
    DoseEquivalentMeasure,
    DoseEquivalentUom,
    DoubleConstantArray,
    DoubleHdf5Array,
    DoubleLatticeArray,
    DoubleLookup,
    DoubleTableLookup,
    DynamicViscosityMeasure,
    DynamicViscosityUom,
    EX_Extent,
    EX_Extent_Type,
    EX_GeographicExtent_PropertyType,
    EX_TemporalExtent,
    EX_TemporalExtent_PropertyType,
    EX_TemporalExtent_Type,
    EX_VerticalExtent,
    EX_VerticalExtent_PropertyType,
    EX_VerticalExtent_Type,
    EarthModelInterpretation,
    EdgePatch,
    Edges,
    ElectricChargeMeasure,
    ElectricChargePerAreaMeasure,
    ElectricChargePerAreaUom,
    ElectricChargePerMassMeasure,
    ElectricChargePerMassUom,
    ElectricChargePerVolumeMeasure,
    ElectricChargePerVolumeUom,
    ElectricChargeUom,
    ElectricConductanceMeasure,
    ElectricConductanceUom,
    ElectricConductivityMeasure,
    ElectricConductivityUom,
    ElectricCurrentDensityMeasure,
    ElectricCurrentDensityUom,
    ElectricCurrentMeasure,
    ElectricCurrentUom,
    ElectricFieldStrengthMeasure,
    ElectricFieldStrengthUom,
    ElectricPotentialDifferenceMeasure,
    ElectricPotentialDifferenceUom,
    ElectricResistanceMeasure,
    ElectricResistancePerLengthMeasure,
    ElectricResistancePerLengthUom,
    ElectricResistanceUom,
    ElectricalResistivityMeasure,
    ElectricalResistivityUom,
    ElectromagneticMomentMeasure,
    ElectromagneticMomentUom,
    ElementIdentity,
    ElementIndices,
    EllipsoidPropertyType,
    EllipsoidType,
    Ellipsoid_1,
    EllipsoidalCSPropertyType,
    EllipsoidalCSType,
    EllipsoidalCS_1,
    EnergyLengthPerAreaMeasure,
    EnergyLengthPerAreaUom,
    EnergyLengthPerTimeAreaTemperatureMeasure,
    EnergyLengthPerTimeAreaTemperatureUom,
    EnergyMeasure,
    EnergyPerAreaMeasure,
    EnergyPerAreaUom,
    EnergyPerLengthMeasure,
    EnergyPerLengthUom,
    EnergyPerMassMeasure,
    EnergyPerMassPerTimeMeasure,
    EnergyPerMassPerTimeUom,
    EnergyPerMassUom,
    EnergyPerVolumeMeasure,
    EnergyPerVolumeUom,
    EnergyUom,
    EpcExternalPartReference,
    Facet,
    FaultInterpretation,
    FaultThrow,
    FloatingPointQuantityParameter,
    FluidBoundaryFeature,
    FluidContact,
    FluidMarker,
    ForceAreaMeasure,
    ForceAreaUom,
    ForceLengthPerLengthMeasure,
    ForceLengthPerLengthUom,
    ForceMeasure,
    ForcePerForceMeasure,
    ForcePerForceUom,
    ForcePerLengthMeasure,
    ForcePerLengthUom,
    ForcePerVolumeMeasure,
    ForcePerVolumeUom,
    ForceUom,
    FrequencyIntervalMeasure,
    FrequencyIntervalUom,
    FrequencyMeasure,
    FrequencyUom,
    FrontierFeature,
    GeneralConversionPropertyType,
    GenericFeatureInterpretation,
    GeneticBoundaryFeature,
    GeneticBoundaryKind,
    Geobody3dShape,
    GeobodyBoundaryInterpretation,
    GeobodyFeature,
    GeobodyInterpretation,
    GeodeticCRS,
    GeodeticCRSPropertyType,
    GeodeticCRSType,
    GeodeticDatumPropertyType,
    GeodeticDatumType,
    GeodeticDatum_1,
    GeologicBoundaryKind,
    GeologicUnitComposition,
    GeologicUnitFeature,
    GeologicUnitInterpretation,
    GeologicUnitInterpretationIndex,
    GeologicUnitMaterialImplacement,
    GlobalChronostratigraphicColumn,
    GmlProjectedCrsDefinition,
    GmlVerticalCrsDefinition,
    GpGridColumnLayerGrid,
    GpGridIjkGridPatch,
    GpGridRepresentation,
    GpGridUnstructuredColumnLayerGridPatch,
    GpGridUnstructuredGridPatch,
    Grid2dPatch,
    Grid2dRepresentation,
    Grid2dSetRepresentation,
    GridConnectionSetRepresentation,
    GridGeometryAttachment,
    Hdf5Dataset,
    HeatCapacityMeasure,
    HeatCapacityUom,
    HeatFlowRateMeasure,
    HeatFlowRateUom,
    HeatTransferCoefficientMeasure,
    HeatTransferCoefficientUom,
    HorizonInterpretation,
    HorizonInterpretationIndex,
    HorizontalPlaneGeometry,
    IdentifiedObjectType,
    IdentityKind,
    IjGaps,
    IjSplitColumnEdges,
    IjkGridGeometry,
    IjkGridRepresentation,
    IjkParentWindow,
    IlluminanceMeasure,
    IlluminanceUom,
    IndexableElements,
    InductanceMeasure,
    InductanceUom,
    IntegerArrayFromBooleanMaskArray,
    IntegerConstantArray,
    IntegerHdf5Array,
    IntegerLatticeArray,
    IntegerQuantityParameter,
    IntegerRangeArray,
    IntervalGridCells,
    IntervalStratigraphicUnits,
    Intervals,
    IsothermalCompressibilityMeasure,
    IsothermalCompressibilityUom,
    KDirection,
    KGaps,
    KinematicViscosityMeasure,
    KinematicViscosityUom,
    LengthMeasure,
    LengthPerLengthMeasure,
    LengthPerLengthUom,
    LengthPerMassMeasure,
    LengthPerMassUom,
    LengthPerPressureMeasure,
    LengthPerPressureUom,
    LengthPerTemperatureMeasure,
    LengthPerTemperatureUom,
    LengthPerTimeMeasure,
    LengthPerTimeUom,
    LengthPerVolumeMeasure,
    LengthPerVolumeUom,
    LengthUom,
    LightExposureMeasure,
    LightExposureUom,
    LineRole,
    LinearAccelerationMeasure,
    LinearAccelerationUom,
    LinearThermalExpansionMeasure,
    LinearThermalExpansionUom,
    LocalDepth3dCrs,
    LocalGridSet,
    LocalPropertyKind,
    LocalTime3dCrs,
    LogarithmicPowerRatioMeasure,
    LogarithmicPowerRatioPerLengthMeasure,
    LogarithmicPowerRatioPerLengthUom,
    LogarithmicPowerRatioUom,
    LuminanceMeasure,
    LuminanceUom,
    LuminousEfficacyMeasure,
    LuminousEfficacyUom,
    LuminousFluxMeasure,
    LuminousFluxUom,
    LuminousIntensityMeasure,
    LuminousIntensityUom,
    MD_Identifier,
    MD_Identifier_PropertyType,
    MD_Identifier_Type,
    MagneticDipoleMomentMeasure,
    MagneticDipoleMomentUom,
    MagneticFieldStrengthMeasure,
    MagneticFieldStrengthUom,
    MagneticFluxDensityMeasure,
    MagneticFluxDensityPerLengthMeasure,
    MagneticFluxDensityPerLengthUom,
    MagneticFluxDensityUom,
    MagneticFluxMeasure,
    MagneticFluxUom,
    MagneticPermeabilityMeasure,
    MagneticPermeabilityUom,
    MagneticVectorPotentialMeasure,
    MagneticVectorPotentialUom,
    MassLengthMeasure,
    MassLengthUom,
    MassMeasure,
    MassPerAreaMeasure,
    MassPerAreaUom,
    MassPerEnergyMeasure,
    MassPerEnergyUom,
    MassPerLengthMeasure,
    MassPerLengthUom,
    MassPerMassMeasure,
    MassPerMassUom,
    MassPerTimeMeasure,
    MassPerTimePerAreaMeasure,
    MassPerTimePerAreaUom,
    MassPerTimePerLengthMeasure,
    MassPerTimePerLengthUom,
    MassPerTimeUom,
    MassPerVolumeMeasure,
    MassPerVolumePerLengthMeasure,
    MassPerVolumePerLengthUom,
    MassPerVolumeUom,
    MassUom,
    MdDatum,
    MdDomain,
    MdReference,
    MobilityMeasure,
    MobilityUom,
    MolarEnergyMeasure,
    MolarEnergyUom,
    MolarHeatCapacityMeasure,
    MolarHeatCapacityUom,
    MolarVolumeMeasure,
    MolarVolumeUom,
    MolecularWeightMeasure,
    MolecularWeightUom,
    MomentOfForceMeasure,
    MomentOfForceUom,
    MomentOfInertiaMeasure,
    MomentOfInertiaUom,
    MomentumMeasure,
    MomentumUom,
    MultipleContactInterpretationPart,
    NameValuePair,
    NilReasonEnumeration_value,
    NodePatch,
    NodesPerCell,
    NonSealedContactRepresentationPart,
    NonSealedSurfaceFrameworkRepresentation,
    NormalizedPowerMeasure,
    NormalizedPowerUom,
    ObjectAlias,
    ObjectParameterKey,
    OrderingCriteria,
    OrganizationFeature,
    OrganizationKind,
    OrientedMacroFace,
    OverlapVolume,
    ParameterKind,
    ParameterTemplate,
    ParametricLineArray,
    ParametricLineFromRepresentationGeometry,
    ParametricLineFromRepresentationLatticeArray,
    ParametricLineGeometry,
    ParametricLineIntersections,
    Patch,
    Patch1d,
    PatchBoundaries,
    PatchOfGeometry,
    PatchOfPoints,
    PatchOfValues,
    PermeabilityLengthMeasure,
    PermeabilityLengthUom,
    PermeabilityRockMeasure,
    PermeabilityRockUom,
    PermittivityMeasure,
    PermittivityUom,
    Phase,
    PillarShape,
    PlaneAngleMeasure,
    PlaneAngleUom,
    PlaneSetRepresentation,
    Point2dHdf5Array,
    Point3d,
    Point3dFromRepresentationLatticeArray,
    Point3dHdf5Array,
    Point3dLatticeArray,
    Point3dOffset,
    Point3dParametricArray,
    Point3dZValueArray,
    PointGeometry,
    PointSetRepresentation,
    PointsProperty,
    PolylineRepresentation,
    PolylineSetPatch,
    PolylineSetRepresentation,
    PotentialDifferencePerPowerDropMeasure,
    PotentialDifferencePerPowerDropUom,
    PowerMeasure,
    PowerPerAreaMeasure,
    PowerPerAreaUom,
    PowerPerPowerMeasure,
    PowerPerPowerUom,
    PowerPerVolumeMeasure,
    PowerPerVolumeUom,
    PowerUom,
    PressureMeasure,
    PressurePerTimeMeasure,
    PressurePerTimeUom,
    PressurePerVolumeMeasure,
    PressurePerVolumeUom,
    PressureSquaredMeasure,
    PressureSquaredPerForceTimePerAreaMeasure,
    PressureSquaredPerForceTimePerAreaUom,
    PressureSquaredUom,
    PressureTimePerVolumeMeasure,
    PressureTimePerVolumeUom,
    PressureUom,
    PrimeMeridianPropertyType,
    PrimeMeridianType,
    PrimeMeridian_1,
    ProjectedCRS,
    ProjectedCRSType,
    ProjectedCrsEpsgCode,
    ProjectedCrs_1,
    ProjectedUnknownCrs,
    PropertyKind,
    PropertyKindFacet,
    PropertySet,
    PropertyValuesPatch,
    QuantityOfLightMeasure,
    QuantityOfLightUom,
    RadianceMeasure,
    RadianceUom,
    RadiantIntensityMeasure,
    RadiantIntensityUom,
    Real,
    Real_PropertyType,
    ReciprocalAreaMeasure,
    ReciprocalAreaUom,
    ReciprocalElectricPotentialDifferenceMeasure,
    ReciprocalElectricPotentialDifferenceUom,
    ReciprocalForceMeasure,
    ReciprocalForceUom,
    ReciprocalLengthMeasure,
    ReciprocalLengthUom,
    ReciprocalMassMeasure,
    ReciprocalMassTimeMeasure,
    ReciprocalMassTimeUom,
    ReciprocalMassUom,
    ReciprocalPressureMeasure,
    ReciprocalPressureUom,
    ReciprocalTimeMeasure,
    ReciprocalTimeUom,
    ReciprocalVolumeMeasure,
    ReciprocalVolumeUom,
    RedefinedGeometryRepresentation,
    ReferenceType,
    Regrid,
    RelatedTimeType,
    RelatedTimeType_relativePosition,
    ReluctanceMeasure,
    ReluctanceUom,
    RepresentationIdentity,
    RepresentationIdentitySet,
    RepresentationSetRepresentation,
    ResqmlJaggedArray,
    ResqmlPropertyKind,
    ResqmlUom,
    RockFluidOrganizationInterpretation,
    RockFluidUnitFeature,
    RockFluidUnitInterpretation,
    RockFluidUnitInterpretationIndex,
    SC_CRS_PropertyType,
    SealedContactRepresentationPart,
    SealedSurfaceFrameworkRepresentation,
    SealedVolumeFrameworkRepresentation,
    SecondDefiningParameter_1,
    SecondMomentOfAreaMeasure,
    SecondMomentOfAreaUom,
    Seismic2dCoordinates,
    Seismic3dCoordinates,
    SeismicLatticeFeature,
    SeismicLatticeSetFeature,
    SeismicLineFeature,
    SeismicLineSetFeature,
    SequenceStratigraphySurface,
    SignalingEventPerTimeMeasure,
    SignalingEventPerTimeUom,
    SolidAngleMeasure,
    SolidAngleUom,
    SpecificHeatCapacityMeasure,
    SpecificHeatCapacityUom,
    SphericalCSPropertyType,
    SphericalCSType,
    SphericalCS_1,
    SplitEdges,
    SplitFaces,
    SplitNodePatch,
    StandardPropertyKind,
    StratigraphicColumn,
    StratigraphicColumnRankInterpretation,
    StratigraphicOccurrenceInterpretation,
    StratigraphicUnitFeature,
    StratigraphicUnitInterpretation,
    StratigraphicUnitInterpretationIndex,
    StreamlineFlux,
    StreamlinePolylineSetPatch,
    StreamlineWellbores,
    StreamlinesFeature,
    StreamlinesRepresentation,
    StringHdf5Array,
    StringLookup,
    StringOrRefType,
    StringParameter,
    StringTableLookup,
    StructuralOrganizationInterpretation,
    SubRepresentation,
    SubRepresentationPatch,
    SubnodeNodeObject,
    SubnodePatch,
    SubnodeTopology,
    SurfaceRole,
    TM_Primitive_PropertyType,
    TectonicBoundaryFeature,
    TectonicBoundaryKind,
    TemperatureIntervalMeasure,
    TemperatureIntervalPerLengthMeasure,
    TemperatureIntervalPerLengthUom,
    TemperatureIntervalPerPressureMeasure,
    TemperatureIntervalPerPressureUom,
    TemperatureIntervalPerTimeMeasure,
    TemperatureIntervalPerTimeUom,
    TemperatureIntervalUom,
    ThermalConductanceMeasure,
    ThermalConductanceUom,
    ThermalConductivityMeasure,
    ThermalConductivityUom,
    ThermalDiffusivityMeasure,
    ThermalDiffusivityUom,
    ThermalInsulanceMeasure,
    ThermalInsulanceUom,
    ThermalResistanceMeasure,
    ThermalResistanceUom,
    ThermodynamicTemperatureMeasure,
    ThermodynamicTemperatureUom,
    ThreePoint3d,
    ThrowKind,
    TiltedPlaneGeometry,
    TimeIndex,
    TimeIndexParameter,
    TimeIndexParameterKey,
    TimeIndices,
    TimeInterval,
    TimeMeasure,
    TimePerLengthMeasure,
    TimePerLengthUom,
    TimePerMassMeasure,
    TimePerMassUom,
    TimePerTimeMeasure,
    TimePerTimeUom,
    TimePerVolumeMeasure,
    TimePerVolumeUom,
    TimePrimitivePropertyType,
    TimeSeries,
    TimeSeriesParentage,
    TimeSetKind,
    TimeUom,
    Timestamp,
    TrianglePatch,
    TriangulatedSetRepresentation,
    TruncatedIjkGridRepresentation,
    TruncatedUnstructuredColumnLayerGridRepresentation,
    TruncationCellPatch,
    URL,
    URL_PropertyType,
    UniformSubnodePatch,
    UnstructuredColumnEdges,
    UnstructuredColumnLayerGridGeometry,
    UnstructuredColumnLayerGridRepresentation,
    UnstructuredGridGeometry,
    UnstructuredGridHingeNodeFaces,
    UnstructuredGridRepresentation,
    UnstructuredSubnodeTopology,
    VariableSubnodePatch,
    VerticalCRS,
    VerticalCRSType,
    VerticalCSPropertyType,
    VerticalCSType,
    VerticalCS_1,
    VerticalCrsEpsgCode,
    VerticalCrs_1,
    VerticalDatumPropertyType,
    VerticalDatumType,
    VerticalDatum_1,
    VerticalDirection,
    VerticalUnknownCrs,
    VolumeFlowRatePerVolumeFlowRateMeasure,
    VolumeFlowRatePerVolumeFlowRateUom,
    VolumeMeasure,
    VolumePerAreaMeasure,
    VolumePerAreaUom,
    VolumePerLengthMeasure,
    VolumePerLengthUom,
    VolumePerMassMeasure,
    VolumePerMassUom,
    VolumePerPressureMeasure,
    VolumePerPressureUom,
    VolumePerRotationMeasure,
    VolumePerRotationUom,
    VolumePerTimeLengthMeasure,
    VolumePerTimeLengthUom,
    VolumePerTimeMeasure,
    VolumePerTimePerAreaMeasure,
    VolumePerTimePerAreaUom,
    VolumePerTimePerLengthMeasure,
    VolumePerTimePerLengthUom,
    VolumePerTimePerPressureLengthMeasure,
    VolumePerTimePerPressureLengthUom,
    VolumePerTimePerPressureMeasure,
    VolumePerTimePerPressureUom,
    VolumePerTimePerTimeMeasure,
    VolumePerTimePerTimeUom,
    VolumePerTimePerVolumeMeasure,
    VolumePerTimePerVolumeUom,
    VolumePerTimeUom,
    VolumePerVolumeMeasure,
    VolumePerVolumeUom,
    VolumeRegion,
    VolumeShell,
    VolumeUom,
    VolumetricHeatTransferCoefficientMeasure,
    VolumetricHeatTransferCoefficientUom,
    VolumetricThermalExpansionMeasure,
    VolumetricThermalExpansionUom,
    WellboreFeature,
    WellboreFrameRepresentation,
    WellboreInterpretation,
    WellboreMarker,
    WellboreMarkerFrameRepresentation,
    WellboreTrajectoryParentIntersection,
    WellboreTrajectoryRepresentation,
    WitsmlWellboreReference,
    actuate_value,
    anchorDefinition,
    axis,
    axisAbbrev,
    axisDirection,
    baseGeodeticCRS,
    cartesianCS_2,
    conversion,
    coordinateOperationAccuracy,
    description,
    descriptionReference,
    domainOfValidity,
    ellipsoid_2,
    ellipsoidalCS_2,
    geodeticDatum_2,
    greenwichLongitude,
    identifier,
    maximumValue,
    minimumValue,
    name,
    obj_Activity,
    obj_ActivityTemplate,
    obj_BlockedWellboreRepresentation,
    obj_BoundaryFeature,
    obj_BoundaryFeatureInterpretation,
    obj_CategoricalProperty,
    obj_CategoricalPropertySeries,
    obj_CommentProperty,
    obj_CommentPropertySeries,
    obj_ContinuousProperty,
    obj_ContinuousPropertySeries,
    obj_DeviationSurveyRepresentation,
    obj_DiscreteProperty,
    obj_DiscretePropertySeries,
    obj_DoubleTableLookup,
    obj_EarthModelInterpretation,
    obj_EpcExternalPartReference,
    obj_FaultInterpretation,
    obj_FluidBoundaryFeature,
    obj_FrontierFeature,
    obj_GenericFeatureInterpretation,
    obj_GeneticBoundaryFeature,
    obj_GeobodyBoundaryInterpretation,
    obj_GeobodyFeature,
    obj_GeobodyInterpretation,
    obj_GeologicUnitFeature,
    obj_GeologicUnitInterpretation,
    obj_GlobalChronostratigraphicColumn,
    obj_GpGridRepresentation,
    obj_Grid2dRepresentation,
    obj_Grid2dSetRepresentation,
    obj_GridConnectionSetRepresentation,
    obj_HorizonInterpretation,
    obj_IjkGridRepresentation,
    obj_LocalDepth3dCrs,
    obj_LocalGridSet,
    obj_LocalTime3dCrs,
    obj_MdDatum,
    obj_NonSealedSurfaceFrameworkRepresentation,
    obj_OrganizationFeature,
    obj_PlaneSetRepresentation,
    obj_PointSetRepresentation,
    obj_PointsProperty,
    obj_PolylineRepresentation,
    obj_PolylineSetRepresentation,
    obj_PropertyKind,
    obj_PropertySet,
    obj_RedefinedGeometryRepresentation,
    obj_RepresentationIdentitySet,
    obj_RepresentationSetRepresentation,
    obj_RockFluidOrganizationInterpretation,
    obj_RockFluidUnitFeature,
    obj_RockFluidUnitInterpretation,
    obj_SealedSurfaceFrameworkRepresentation,
    obj_SealedVolumeFrameworkRepresentation,
    obj_SeismicLatticeFeature,
    obj_SeismicLineFeature,
    obj_SeismicLineSetFeature,
    obj_StratigraphicColumn,
    obj_StratigraphicColumnRankInterpretation,
    obj_StratigraphicOccurrenceInterpretation,
    obj_StratigraphicUnitFeature,
    obj_StratigraphicUnitInterpretation,
    obj_StreamlinesFeature,
    obj_StreamlinesRepresentation,
    obj_StringTableLookup,
    obj_StructuralOrganizationInterpretation,
    obj_SubRepresentation,
    obj_TectonicBoundaryFeature,
    obj_TimeSeries,
    obj_TriangulatedSetRepresentation,
    obj_TruncatedIjkGridRepresentation,
    obj_TruncatedUnstructuredColumnLayerGridRepresentation,
    obj_UnstructuredColumnLayerGridRepresentation,
    obj_UnstructuredGridRepresentation,
    obj_WellboreFeature,
    obj_WellboreFrameRepresentation,
    obj_WellboreInterpretation,
    obj_WellboreMarkerFrameRepresentation,
    obj_WellboreTrajectoryRepresentation,
    operationVersion,
    primeMeridian_2,
    rangeMeaning,
    realizationEpoch,
    remarks,
    scope,
    secondDefiningParameter_2,
    semiMajorAxis,
    show_value,
    sourceCRS,
    sphericalCS_2,
    targetCRS,
    verticalCS_2,
    verticalDatum_2,
)

__all__ = [
    "APIGammaRayMeasure",
    "APIGammaRayUom",
    "APIGravityMeasure",
    "APIGravityUom",
    "APINeutronMeasure",
    "APINeutronUom",
    "AbsorbedDoseMeasure",
    "AbsorbedDoseUom",
    "AbstractActivityParameter",
    "AbstractBooleanArray",
    "AbstractCRSType",
    "AbstractCitedDataObject",
    "AbstractColumnLayerGridGeometry",
    "AbstractColumnLayerGridRepresentation",
    "AbstractContactInterpretationPart",
    "AbstractContactRepresentationPart",
    "AbstractContextualObject",
    "AbstractCoordinateOperationType",
    "AbstractCoordinateSystemType",
    "AbstractDQ_Element_Type",
    "AbstractDQ_PositionalAccuracy_Type",
    "AbstractDQ_Result_Type",
    "AbstractDataObject",
    "AbstractDatumType",
    "AbstractDoubleArray",
    "AbstractEX_GeographicExtent_Type",
    "AbstractFeature",
    "AbstractFeatureInterpretation",
    "AbstractGMLType",
    "AbstractGeneralConversionType",
    "AbstractGeneralDerivedCRSType",
    "AbstractGeologicFeature",
    "AbstractGeometry",
    "AbstractGridGeometry",
    "AbstractGridRepresentation",
    "AbstractIntegerArray",
    "AbstractLocal3dCrs",
    "AbstractObject",
    "AbstractObject_Type",
    "AbstractOrganizationInterpretation",
    "AbstractParameterKey",
    "AbstractParametricLineArray",
    "AbstractParametricLineGeometry",
    "AbstractParentWindow",
    "AbstractPlaneGeometry",
    "AbstractPoint3dArray",
    "AbstractProjectedCrs",
    "AbstractProperty",
    "AbstractPropertyKind",
    "AbstractPropertyLookup",
    "AbstractRepresentation",
    "AbstractResqmlDataObject",
    "AbstractSeismicCoordinates",
    "AbstractSeismicSurveyFeature",
    "AbstractStratigraphicOrganizationInterpretation",
    "AbstractSurfaceFrameworkRepresentation",
    "AbstractSurfaceRepresentation",
    "AbstractTechnicalFeature",
    "AbstractTimeObjectType",
    "AbstractTimePrimitiveType",
    "AbstractTruncatedColumnLayerGridRepresentation",
    "AbstractValueArray",
    "AbstractValuesProperty",
    "AbstractVerticalCrs",
    "Activation",
    "Activity",
    "ActivityOfRadioactivityMeasure",
    "ActivityOfRadioactivityUom",
    "ActivityTemplate",
    "AdditionalGridPoints",
    "AdditionalGridTopology",
    "AggregationType",
    "AmountOfSubstanceMeasure",
    "AmountOfSubstancePerAmountOfSubstanceMeasure",
    "AmountOfSubstancePerAmountOfSubstanceUom",
    "AmountOfSubstancePerAreaMeasure",
    "AmountOfSubstancePerAreaUom",
    "AmountOfSubstancePerTimeMeasure",
    "AmountOfSubstancePerTimePerAreaMeasure",
    "AmountOfSubstancePerTimePerAreaUom",
    "AmountOfSubstancePerTimeUom",
    "AmountOfSubstancePerVolumeMeasure",
    "AmountOfSubstancePerVolumeUom",
    "AmountOfSubstanceUom",
    "AnglePerLengthMeasure",
    "AnglePerLengthUom",
    "AnglePerVolumeMeasure",
    "AnglePerVolumeUom",
    "AngularAccelerationMeasure",
    "AngularAccelerationUom",
    "AngularVelocityMeasure",
    "AngularVelocityUom",
    "AreaMeasure",
    "AreaPerAmountOfSubstanceMeasure",
    "AreaPerAmountOfSubstanceUom",
    "AreaPerAreaMeasure",
    "AreaPerAreaUom",
    "AreaPerMassMeasure",
    "AreaPerMassUom",
    "AreaPerTimeMeasure",
    "AreaPerTimeUom",
    "AreaPerVolumeMeasure",
    "AreaPerVolumeUom",
    "AreaUom",
    "AttenuationPerFrequencyIntervalMeasure",
    "AttenuationPerFrequencyIntervalUom",
    "AxisOrder2d",
    "BinaryContactInterpretationPart",
    "BlockedWellboreRepresentation",
    "Boolean",
    "BooleanArrayFromDiscretePropertyArray",
    "BooleanArrayFromIndexArray",
    "BooleanConstantArray",
    "BooleanHdf5Array",
    "Boolean_PropertyType",
    "BoundaryFeature",
    "BoundaryFeatureInterpretation",
    "BoundaryRelation",
    "CI_Address",
    "CI_Address_PropertyType",
    "CI_Address_Type",
    "CI_Citation",
    "CI_Citation_PropertyType",
    "CI_Citation_Type",
    "CI_Contact",
    "CI_Contact_PropertyType",
    "CI_Contact_Type",
    "CI_Date",
    "CI_DateTypeCode",
    "CI_DateTypeCode_PropertyType",
    "CI_Date_PropertyType",
    "CI_Date_Type",
    "CI_OnLineFunctionCode",
    "CI_OnLineFunctionCode_PropertyType",
    "CI_OnlineResource",
    "CI_OnlineResource_PropertyType",
    "CI_OnlineResource_Type",
    "CI_PresentationFormCode",
    "CI_PresentationFormCode_PropertyType",
    "CI_ResponsibleParty",
    "CI_ResponsibleParty_PropertyType",
    "CI_ResponsibleParty_Type",
    "CI_RoleCode",
    "CI_RoleCode_PropertyType",
    "CI_Series",
    "CI_Series_PropertyType",
    "CI_Series_Type",
    "CI_Telephone",
    "CI_Telephone_PropertyType",
    "CI_Telephone_Type",
    "CRSPropertyType",
    "CapacitanceMeasure",
    "CapacitanceUom",
    "CartesianCSPropertyType",
    "CartesianCSType",
    "CartesianCS_1",
    "CategoricalProperty",
    "CategoricalPropertySeries",
    "CellFluidPhaseUnits",
    "CellOverlap",
    "CellParentWindow",
    "CellShape",
    "CellStratigraphicUnits",
    "CharacterString",
    "CharacterString_PropertyType",
    "ChronostratigraphicRank",
    "Citation",
    "CodeListValue_Type",
    "CodeType",
    "CodeWithAuthorityType",
    "ColumnLayerParentWindow",
    "ColumnLayerSplitColumnEdges",
    "ColumnLayerSplitCoordinateLines",
    "ColumnLayerSubnodeTopology",
    "ColumnShape",
    "ColumnSubnodePatch",
    "CommentProperty",
    "CommentPropertySeries",
    "ConnectionInterpretations",
    "ContactElementReference",
    "ContactIdentity",
    "ContactMode",
    "ContactPatch",
    "ContactRelationship",
    "ContactRepresentationReference",
    "ContactSide",
    "ContactVerb",
    "ContinuousProperty",
    "ContinuousPropertySeries",
    "CoordinateSystemAxis",
    "CoordinateSystemAxisPropertyType",
    "CoordinateSystemAxisType",
    "CustomData",
    "DQ_EvaluationMethodTypeCode",
    "DQ_EvaluationMethodTypeCode_PropertyType",
    "DQ_Result_PropertyType",
    "DataObjectParameter",
    "DataObjectReference",
    "DataTransferSpeedMeasure",
    "DataTransferSpeedUom",
    "Date",
    "DateTime",
    "DateTime_PropertyType",
    "Date_PropertyType",
    "Definition",
    "DefinitionBaseType",
    "DefinitionType",
    "DepositionMode",
    "DeviationSurveyRepresentation",
    "DiffusionCoefficientMeasure",
    "DiffusionCoefficientUom",
    "DigitalStorageMeasure",
    "DigitalStorageUom",
    "DimensionlessMeasure",
    "DimensionlessUom",
    "DipoleMomentMeasure",
    "DipoleMomentUom",
    "DiscreteProperty",
    "DiscretePropertySeries",
    "Domain",
    "DoseEquivalentMeasure",
    "DoseEquivalentUom",
    "DoubleConstantArray",
    "DoubleHdf5Array",
    "DoubleLatticeArray",
    "DoubleLookup",
    "DoubleTableLookup",
    "DynamicViscosityMeasure",
    "DynamicViscosityUom",
    "EX_Extent",
    "EX_Extent_Type",
    "EX_GeographicExtent_PropertyType",
    "EX_TemporalExtent",
    "EX_TemporalExtent_PropertyType",
    "EX_TemporalExtent_Type",
    "EX_VerticalExtent",
    "EX_VerticalExtent_PropertyType",
    "EX_VerticalExtent_Type",
    "EarthModelInterpretation",
    "EdgePatch",
    "Edges",
    "ElectricChargeMeasure",
    "ElectricChargePerAreaMeasure",
    "ElectricChargePerAreaUom",
    "ElectricChargePerMassMeasure",
    "ElectricChargePerMassUom",
    "ElectricChargePerVolumeMeasure",
    "ElectricChargePerVolumeUom",
    "ElectricChargeUom",
    "ElectricConductanceMeasure",
    "ElectricConductanceUom",
    "ElectricConductivityMeasure",
    "ElectricConductivityUom",
    "ElectricCurrentDensityMeasure",
    "ElectricCurrentDensityUom",
    "ElectricCurrentMeasure",
    "ElectricCurrentUom",
    "ElectricFieldStrengthMeasure",
    "ElectricFieldStrengthUom",
    "ElectricPotentialDifferenceMeasure",
    "ElectricPotentialDifferenceUom",
    "ElectricResistanceMeasure",
    "ElectricResistancePerLengthMeasure",
    "ElectricResistancePerLengthUom",
    "ElectricResistanceUom",
    "ElectricalResistivityMeasure",
    "ElectricalResistivityUom",
    "ElectromagneticMomentMeasure",
    "ElectromagneticMomentUom",
    "ElementIdentity",
    "ElementIndices",
    "EllipsoidPropertyType",
    "EllipsoidType",
    "Ellipsoid_1",
    "EllipsoidalCSPropertyType",
    "EllipsoidalCSType",
    "EllipsoidalCS_1",
    "EnergyLengthPerAreaMeasure",
    "EnergyLengthPerAreaUom",
    "EnergyLengthPerTimeAreaTemperatureMeasure",
    "EnergyLengthPerTimeAreaTemperatureUom",
    "EnergyMeasure",
    "EnergyPerAreaMeasure",
    "EnergyPerAreaUom",
    "EnergyPerLengthMeasure",
    "EnergyPerLengthUom",
    "EnergyPerMassMeasure",
    "EnergyPerMassPerTimeMeasure",
    "EnergyPerMassPerTimeUom",
    "EnergyPerMassUom",
    "EnergyPerVolumeMeasure",
    "EnergyPerVolumeUom",
    "EnergyUom",
    "EpcExternalPartReference",
    "Facet",
    "FaultInterpretation",
    "FaultThrow",
    "FloatingPointQuantityParameter",
    "FluidBoundaryFeature",
    "FluidContact",
    "FluidMarker",
    "ForceAreaMeasure",
    "ForceAreaUom",
    "ForceLengthPerLengthMeasure",
    "ForceLengthPerLengthUom",
    "ForceMeasure",
    "ForcePerForceMeasure",
    "ForcePerForceUom",
    "ForcePerLengthMeasure",
    "ForcePerLengthUom",
    "ForcePerVolumeMeasure",
    "ForcePerVolumeUom",
    "ForceUom",
    "FrequencyIntervalMeasure",
    "FrequencyIntervalUom",
    "FrequencyMeasure",
    "FrequencyUom",
    "FrontierFeature",
    "GeneralConversionPropertyType",
    "GenericFeatureInterpretation",
    "GeneticBoundaryFeature",
    "GeneticBoundaryKind",
    "Geobody3dShape",
    "GeobodyBoundaryInterpretation",
    "GeobodyFeature",
    "GeobodyInterpretation",
    "GeodeticCRS",
    "GeodeticCRSPropertyType",
    "GeodeticCRSType",
    "GeodeticDatumPropertyType",
    "GeodeticDatumType",
    "GeodeticDatum_1",
    "GeologicBoundaryKind",
    "GeologicUnitComposition",
    "GeologicUnitFeature",
    "GeologicUnitInterpretation",
    "GeologicUnitInterpretationIndex",
    "GeologicUnitMaterialImplacement",
    "GlobalChronostratigraphicColumn",
    "GmlProjectedCrsDefinition",
    "GmlVerticalCrsDefinition",
    "GpGridColumnLayerGrid",
    "GpGridIjkGridPatch",
    "GpGridRepresentation",
    "GpGridUnstructuredColumnLayerGridPatch",
    "GpGridUnstructuredGridPatch",
    "Grid2dPatch",
    "Grid2dRepresentation",
    "Grid2dSetRepresentation",
    "GridConnectionSetRepresentation",
    "GridGeometryAttachment",
    "Hdf5Dataset",
    "HeatCapacityMeasure",
    "HeatCapacityUom",
    "HeatFlowRateMeasure",
    "HeatFlowRateUom",
    "HeatTransferCoefficientMeasure",
    "HeatTransferCoefficientUom",
    "HorizonInterpretation",
    "HorizonInterpretationIndex",
    "HorizontalPlaneGeometry",
    "IdentifiedObjectType",
    "IdentityKind",
    "IjGaps",
    "IjSplitColumnEdges",
    "IjkGridGeometry",
    "IjkGridRepresentation",
    "IjkParentWindow",
    "IlluminanceMeasure",
    "IlluminanceUom",
    "IndexableElements",
    "InductanceMeasure",
    "InductanceUom",
    "IntegerArrayFromBooleanMaskArray",
    "IntegerConstantArray",
    "IntegerHdf5Array",
    "IntegerLatticeArray",
    "IntegerQuantityParameter",
    "IntegerRangeArray",
    "IntervalGridCells",
    "IntervalStratigraphicUnits",
    "Intervals",
    "IsothermalCompressibilityMeasure",
    "IsothermalCompressibilityUom",
    "KDirection",
    "KGaps",
    "KinematicViscosityMeasure",
    "KinematicViscosityUom",
    "LengthMeasure",
    "LengthPerLengthMeasure",
    "LengthPerLengthUom",
    "LengthPerMassMeasure",
    "LengthPerMassUom",
    "LengthPerPressureMeasure",
    "LengthPerPressureUom",
    "LengthPerTemperatureMeasure",
    "LengthPerTemperatureUom",
    "LengthPerTimeMeasure",
    "LengthPerTimeUom",
    "LengthPerVolumeMeasure",
    "LengthPerVolumeUom",
    "LengthUom",
    "LightExposureMeasure",
    "LightExposureUom",
    "LineRole",
    "LinearAccelerationMeasure",
    "LinearAccelerationUom",
    "LinearThermalExpansionMeasure",
    "LinearThermalExpansionUom",
    "LocalDepth3dCrs",
    "LocalGridSet",
    "LocalPropertyKind",
    "LocalTime3dCrs",
    "LogarithmicPowerRatioMeasure",
    "LogarithmicPowerRatioPerLengthMeasure",
    "LogarithmicPowerRatioPerLengthUom",
    "LogarithmicPowerRatioUom",
    "LuminanceMeasure",
    "LuminanceUom",
    "LuminousEfficacyMeasure",
    "LuminousEfficacyUom",
    "LuminousFluxMeasure",
    "LuminousFluxUom",
    "LuminousIntensityMeasure",
    "LuminousIntensityUom",
    "MD_Identifier",
    "MD_Identifier_PropertyType",
    "MD_Identifier_Type",
    "MagneticDipoleMomentMeasure",
    "MagneticDipoleMomentUom",
    "MagneticFieldStrengthMeasure",
    "MagneticFieldStrengthUom",
    "MagneticFluxDensityMeasure",
    "MagneticFluxDensityPerLengthMeasure",
    "MagneticFluxDensityPerLengthUom",
    "MagneticFluxDensityUom",
    "MagneticFluxMeasure",
    "MagneticFluxUom",
    "MagneticPermeabilityMeasure",
    "MagneticPermeabilityUom",
    "MagneticVectorPotentialMeasure",
    "MagneticVectorPotentialUom",
    "MassLengthMeasure",
    "MassLengthUom",
    "MassMeasure",
    "MassPerAreaMeasure",
    "MassPerAreaUom",
    "MassPerEnergyMeasure",
    "MassPerEnergyUom",
    "MassPerLengthMeasure",
    "MassPerLengthUom",
    "MassPerMassMeasure",
    "MassPerMassUom",
    "MassPerTimeMeasure",
    "MassPerTimePerAreaMeasure",
    "MassPerTimePerAreaUom",
    "MassPerTimePerLengthMeasure",
    "MassPerTimePerLengthUom",
    "MassPerTimeUom",
    "MassPerVolumeMeasure",
    "MassPerVolumePerLengthMeasure",
    "MassPerVolumePerLengthUom",
    "MassPerVolumeUom",
    "MassUom",
    "MdDatum",
    "MdDomain",
    "MdReference",
    "MobilityMeasure",
    "MobilityUom",
    "MolarEnergyMeasure",
    "MolarEnergyUom",
    "MolarHeatCapacityMeasure",
    "MolarHeatCapacityUom",
    "MolarVolumeMeasure",
    "MolarVolumeUom",
    "MolecularWeightMeasure",
    "MolecularWeightUom",
    "MomentOfForceMeasure",
    "MomentOfForceUom",
    "MomentOfInertiaMeasure",
    "MomentOfInertiaUom",
    "MomentumMeasure",
    "MomentumUom",
    "MultipleContactInterpretationPart",
    "NameValuePair",
    "NilReasonEnumeration_value",
    "NodePatch",
    "NodesPerCell",
    "NonSealedContactRepresentationPart",
    "NonSealedSurfaceFrameworkRepresentation",
    "NormalizedPowerMeasure",
    "NormalizedPowerUom",
    "ObjectAlias",
    "ObjectParameterKey",
    "OrderingCriteria",
    "OrganizationFeature",
    "OrganizationKind",
    "OrientedMacroFace",
    "OverlapVolume",
    "ParameterKind",
    "ParameterTemplate",
    "ParametricLineArray",
    "ParametricLineFromRepresentationGeometry",
    "ParametricLineFromRepresentationLatticeArray",
    "ParametricLineGeometry",
    "ParametricLineIntersections",
    "Patch",
    "Patch1d",
    "PatchBoundaries",
    "PatchOfGeometry",
    "PatchOfPoints",
    "PatchOfValues",
    "PermeabilityLengthMeasure",
    "PermeabilityLengthUom",
    "PermeabilityRockMeasure",
    "PermeabilityRockUom",
    "PermittivityMeasure",
    "PermittivityUom",
    "Phase",
    "PillarShape",
    "PlaneAngleMeasure",
    "PlaneAngleUom",
    "PlaneSetRepresentation",
    "Point2dHdf5Array",
    "Point3d",
    "Point3dFromRepresentationLatticeArray",
    "Point3dHdf5Array",
    "Point3dLatticeArray",
    "Point3dOffset",
    "Point3dParametricArray",
    "Point3dZValueArray",
    "PointGeometry",
    "PointSetRepresentation",
    "PointsProperty",
    "PolylineRepresentation",
    "PolylineSetPatch",
    "PolylineSetRepresentation",
    "PotentialDifferencePerPowerDropMeasure",
    "PotentialDifferencePerPowerDropUom",
    "PowerMeasure",
    "PowerPerAreaMeasure",
    "PowerPerAreaUom",
    "PowerPerPowerMeasure",
    "PowerPerPowerUom",
    "PowerPerVolumeMeasure",
    "PowerPerVolumeUom",
    "PowerUom",
    "PressureMeasure",
    "PressurePerTimeMeasure",
    "PressurePerTimeUom",
    "PressurePerVolumeMeasure",
    "PressurePerVolumeUom",
    "PressureSquaredMeasure",
    "PressureSquaredPerForceTimePerAreaMeasure",
    "PressureSquaredPerForceTimePerAreaUom",
    "PressureSquaredUom",
    "PressureTimePerVolumeMeasure",
    "PressureTimePerVolumeUom",
    "PressureUom",
    "PrimeMeridianPropertyType",
    "PrimeMeridianType",
    "PrimeMeridian_1",
    "ProjectedCRS",
    "ProjectedCRSType",
    "ProjectedCrsEpsgCode",
    "ProjectedCrs_1",
    "ProjectedUnknownCrs",
    "PropertyKind",
    "PropertyKindFacet",
    "PropertySet",
    "PropertyValuesPatch",
    "QuantityOfLightMeasure",
    "QuantityOfLightUom",
    "RadianceMeasure",
    "RadianceUom",
    "RadiantIntensityMeasure",
    "RadiantIntensityUom",
    "Real",
    "Real_PropertyType",
    "ReciprocalAreaMeasure",
    "ReciprocalAreaUom",
    "ReciprocalElectricPotentialDifferenceMeasure",
    "ReciprocalElectricPotentialDifferenceUom",
    "ReciprocalForceMeasure",
    "ReciprocalForceUom",
    "ReciprocalLengthMeasure",
    "ReciprocalLengthUom",
    "ReciprocalMassMeasure",
    "ReciprocalMassTimeMeasure",
    "ReciprocalMassTimeUom",
    "ReciprocalMassUom",
    "ReciprocalPressureMeasure",
    "ReciprocalPressureUom",
    "ReciprocalTimeMeasure",
    "ReciprocalTimeUom",
    "ReciprocalVolumeMeasure",
    "ReciprocalVolumeUom",
    "RedefinedGeometryRepresentation",
    "ReferenceType",
    "Regrid",
    "RelatedTimeType",
    "RelatedTimeType_relativePosition",
    "ReluctanceMeasure",
    "ReluctanceUom",
    "RepresentationIdentity",
    "RepresentationIdentitySet",
    "RepresentationSetRepresentation",
    "ResqmlJaggedArray",
    "ResqmlPropertyKind",
    "ResqmlUom",
    "RockFluidOrganizationInterpretation",
    "RockFluidUnitFeature",
    "RockFluidUnitInterpretation",
    "RockFluidUnitInterpretationIndex",
    "SC_CRS_PropertyType",
    "SealedContactRepresentationPart",
    "SealedSurfaceFrameworkRepresentation",
    "SealedVolumeFrameworkRepresentation",
    "SecondDefiningParameter_1",
    "SecondMomentOfAreaMeasure",
    "SecondMomentOfAreaUom",
    "Seismic2dCoordinates",
    "Seismic3dCoordinates",
    "SeismicLatticeFeature",
    "SeismicLatticeSetFeature",
    "SeismicLineFeature",
    "SeismicLineSetFeature",
    "SequenceStratigraphySurface",
    "SignalingEventPerTimeMeasure",
    "SignalingEventPerTimeUom",
    "SolidAngleMeasure",
    "SolidAngleUom",
    "SpecificHeatCapacityMeasure",
    "SpecificHeatCapacityUom",
    "SphericalCSPropertyType",
    "SphericalCSType",
    "SphericalCS_1",
    "SplitEdges",
    "SplitFaces",
    "SplitNodePatch",
    "StandardPropertyKind",
    "StratigraphicColumn",
    "StratigraphicColumnRankInterpretation",
    "StratigraphicOccurrenceInterpretation",
    "StratigraphicUnitFeature",
    "StratigraphicUnitInterpretation",
    "StratigraphicUnitInterpretationIndex",
    "StreamlineFlux",
    "StreamlinePolylineSetPatch",
    "StreamlineWellbores",
    "StreamlinesFeature",
    "StreamlinesRepresentation",
    "StringHdf5Array",
    "StringLookup",
    "StringOrRefType",
    "StringParameter",
    "StringTableLookup",
    "StructuralOrganizationInterpretation",
    "SubRepresentation",
    "SubRepresentationPatch",
    "SubnodeNodeObject",
    "SubnodePatch",
    "SubnodeTopology",
    "SurfaceRole",
    "TM_Primitive_PropertyType",
    "TectonicBoundaryFeature",
    "TectonicBoundaryKind",
    "TemperatureIntervalMeasure",
    "TemperatureIntervalPerLengthMeasure",
    "TemperatureIntervalPerLengthUom",
    "TemperatureIntervalPerPressureMeasure",
    "TemperatureIntervalPerPressureUom",
    "TemperatureIntervalPerTimeMeasure",
    "TemperatureIntervalPerTimeUom",
    "TemperatureIntervalUom",
    "ThermalConductanceMeasure",
    "ThermalConductanceUom",
    "ThermalConductivityMeasure",
    "ThermalConductivityUom",
    "ThermalDiffusivityMeasure",
    "ThermalDiffusivityUom",
    "ThermalInsulanceMeasure",
    "ThermalInsulanceUom",
    "ThermalResistanceMeasure",
    "ThermalResistanceUom",
    "ThermodynamicTemperatureMeasure",
    "ThermodynamicTemperatureUom",
    "ThreePoint3d",
    "ThrowKind",
    "TiltedPlaneGeometry",
    "TimeIndex",
    "TimeIndexParameter",
    "TimeIndexParameterKey",
    "TimeIndices",
    "TimeInterval",
    "TimeMeasure",
    "TimePerLengthMeasure",
    "TimePerLengthUom",
    "TimePerMassMeasure",
    "TimePerMassUom",
    "TimePerTimeMeasure",
    "TimePerTimeUom",
    "TimePerVolumeMeasure",
    "TimePerVolumeUom",
    "TimePrimitivePropertyType",
    "TimeSeries",
    "TimeSeriesParentage",
    "TimeSetKind",
    "TimeUom",
    "Timestamp",
    "TrianglePatch",
    "TriangulatedSetRepresentation",
    "TruncatedIjkGridRepresentation",
    "TruncatedUnstructuredColumnLayerGridRepresentation",
    "TruncationCellPatch",
    "URL",
    "URL_PropertyType",
    "UniformSubnodePatch",
    "UnstructuredColumnEdges",
    "UnstructuredColumnLayerGridGeometry",
    "UnstructuredColumnLayerGridRepresentation",
    "UnstructuredGridGeometry",
    "UnstructuredGridHingeNodeFaces",
    "UnstructuredGridRepresentation",
    "UnstructuredSubnodeTopology",
    "VariableSubnodePatch",
    "VerticalCRS",
    "VerticalCRSType",
    "VerticalCSPropertyType",
    "VerticalCSType",
    "VerticalCS_1",
    "VerticalCrsEpsgCode",
    "VerticalCrs_1",
    "VerticalDatumPropertyType",
    "VerticalDatumType",
    "VerticalDatum_1",
    "VerticalDirection",
    "VerticalUnknownCrs",
    "VolumeFlowRatePerVolumeFlowRateMeasure",
    "VolumeFlowRatePerVolumeFlowRateUom",
    "VolumeMeasure",
    "VolumePerAreaMeasure",
    "VolumePerAreaUom",
    "VolumePerLengthMeasure",
    "VolumePerLengthUom",
    "VolumePerMassMeasure",
    "VolumePerMassUom",
    "VolumePerPressureMeasure",
    "VolumePerPressureUom",
    "VolumePerRotationMeasure",
    "VolumePerRotationUom",
    "VolumePerTimeLengthMeasure",
    "VolumePerTimeLengthUom",
    "VolumePerTimeMeasure",
    "VolumePerTimePerAreaMeasure",
    "VolumePerTimePerAreaUom",
    "VolumePerTimePerLengthMeasure",
    "VolumePerTimePerLengthUom",
    "VolumePerTimePerPressureLengthMeasure",
    "VolumePerTimePerPressureLengthUom",
    "VolumePerTimePerPressureMeasure",
    "VolumePerTimePerPressureUom",
    "VolumePerTimePerTimeMeasure",
    "VolumePerTimePerTimeUom",
    "VolumePerTimePerVolumeMeasure",
    "VolumePerTimePerVolumeUom",
    "VolumePerTimeUom",
    "VolumePerVolumeMeasure",
    "VolumePerVolumeUom",
    "VolumeRegion",
    "VolumeShell",
    "VolumeUom",
    "VolumetricHeatTransferCoefficientMeasure",
    "VolumetricHeatTransferCoefficientUom",
    "VolumetricThermalExpansionMeasure",
    "VolumetricThermalExpansionUom",
    "WellboreFeature",
    "WellboreFrameRepresentation",
    "WellboreInterpretation",
    "WellboreMarker",
    "WellboreMarkerFrameRepresentation",
    "WellboreTrajectoryParentIntersection",
    "WellboreTrajectoryRepresentation",
    "WitsmlWellboreReference",
    "actuate_value",
    "anchorDefinition",
    "axis",
    "axisAbbrev",
    "axisDirection",
    "baseGeodeticCRS",
    "cartesianCS_2",
    "conversion",
    "coordinateOperationAccuracy",
    "description",
    "descriptionReference",
    "domainOfValidity",
    "ellipsoid_2",
    "ellipsoidalCS_2",
    "geodeticDatum_2",
    "greenwichLongitude",
    "identifier",
    "maximumValue",
    "minimumValue",
    "name",
    "obj_Activity",
    "obj_ActivityTemplate",
    "obj_BlockedWellboreRepresentation",
    "obj_BoundaryFeature",
    "obj_BoundaryFeatureInterpretation",
    "obj_CategoricalProperty",
    "obj_CategoricalPropertySeries",
    "obj_CommentProperty",
    "obj_CommentPropertySeries",
    "obj_ContinuousProperty",
    "obj_ContinuousPropertySeries",
    "obj_DeviationSurveyRepresentation",
    "obj_DiscreteProperty",
    "obj_DiscretePropertySeries",
    "obj_DoubleTableLookup",
    "obj_EarthModelInterpretation",
    "obj_EpcExternalPartReference",
    "obj_FaultInterpretation",
    "obj_FluidBoundaryFeature",
    "obj_FrontierFeature",
    "obj_GenericFeatureInterpretation",
    "obj_GeneticBoundaryFeature",
    "obj_GeobodyBoundaryInterpretation",
    "obj_GeobodyFeature",
    "obj_GeobodyInterpretation",
    "obj_GeologicUnitFeature",
    "obj_GeologicUnitInterpretation",
    "obj_GlobalChronostratigraphicColumn",
    "obj_GpGridRepresentation",
    "obj_Grid2dRepresentation",
    "obj_Grid2dSetRepresentation",
    "obj_GridConnectionSetRepresentation",
    "obj_HorizonInterpretation",
    "obj_IjkGridRepresentation",
    "obj_LocalDepth3dCrs",
    "obj_LocalGridSet",
    "obj_LocalTime3dCrs",
    "obj_MdDatum",
    "obj_NonSealedSurfaceFrameworkRepresentation",
    "obj_OrganizationFeature",
    "obj_PlaneSetRepresentation",
    "obj_PointSetRepresentation",
    "obj_PointsProperty",
    "obj_PolylineRepresentation",
    "obj_PolylineSetRepresentation",
    "obj_PropertyKind",
    "obj_PropertySet",
    "obj_RedefinedGeometryRepresentation",
    "obj_RepresentationIdentitySet",
    "obj_RepresentationSetRepresentation",
    "obj_RockFluidOrganizationInterpretation",
    "obj_RockFluidUnitFeature",
    "obj_RockFluidUnitInterpretation",
    "obj_SealedSurfaceFrameworkRepresentation",
    "obj_SealedVolumeFrameworkRepresentation",
    "obj_SeismicLatticeFeature",
    "obj_SeismicLineFeature",
    "obj_SeismicLineSetFeature",
    "obj_StratigraphicColumn",
    "obj_StratigraphicColumnRankInterpretation",
    "obj_StratigraphicOccurrenceInterpretation",
    "obj_StratigraphicUnitFeature",
    "obj_StratigraphicUnitInterpretation",
    "obj_StreamlinesFeature",
    "obj_StreamlinesRepresentation",
    "obj_StringTableLookup",
    "obj_StructuralOrganizationInterpretation",
    "obj_SubRepresentation",
    "obj_TectonicBoundaryFeature",
    "obj_TimeSeries",
    "obj_TriangulatedSetRepresentation",
    "obj_TruncatedIjkGridRepresentation",
    "obj_TruncatedUnstructuredColumnLayerGridRepresentation",
    "obj_UnstructuredColumnLayerGridRepresentation",
    "obj_UnstructuredGridRepresentation",
    "obj_WellboreFeature",
    "obj_WellboreFrameRepresentation",
    "obj_WellboreInterpretation",
    "obj_WellboreMarkerFrameRepresentation",
    "obj_WellboreTrajectoryRepresentation",
    "operationVersion",
    "primeMeridian_2",
    "rangeMeaning",
    "realizationEpoch",
    "remarks",
    "scope",
    "secondDefiningParameter_2",
    "semiMajorAxis",
    "show_value",
    "sourceCRS",
    "sphericalCS_2",
    "targetCRS",
    "verticalCS_2",
    "verticalDatum_2",
]



import asyncio
import sys
import datetime
import logging
import time
import typing as T
import uuid
from collections import defaultdict
from types import TracebackType
from async_timeout import timeout
import numpy as np
from pydantic import SecretStr
from scipy import interpolate
import websockets
from etpproto.connection import (CommunicationProtocol, ConnectionType,
                                 ETPConnection)
from etpproto.messages import Message, MessageFlags
from xtgeo import RegularSurface
import xtgeo
from scipy.interpolate import griddata
import pyetp.resqml_objects as ro
from pyetp.config import SETTINGS

from pyetp import utils_arrays, utils_xml
from pyetp.types import *
from pyetp.uri import DataObjectURI, DataspaceURI

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


MAXPAYLOADSIZE = 10_000_000  # 10MB


class ETPError(Exception):
    def __init__(self, message: str, code: int):
        self.message = message
        self.code = code
        super().__init__(f"{message} ({code=:})")

    @classmethod
    def from_proto(cls, msg: ProtocolException):
        assert msg.error is not None or msg.errors is not None, "passed no error info"
        error = msg.error or list(msg.errors.values())[0]
        return cls(error.message, error.code)


def get_all_etp_protocol_classes():
    """Update protocol - all exception protocols are now per message"""

    pddict = ETPConnection.generic_transition_table
    pexec = ETPConnection.generic_transition_table["0"]["1000"]  # protocol exception

    for v in pddict.values():
        v["1000"] = pexec

    return pddict


class ETPClient(ETPConnection):

    generic_transition_table = get_all_etp_protocol_classes()

    _recv_events: T.Dict[int, asyncio.Event]
    _recv_buffer: T.Dict[int, T.List[ETPModel]]

    def __init__(self, ws: websockets.WebSocketClientProtocol, default_dataspace_uri: T.Union[DataspaceURI , None], timeout=10.):
        super().__init__(connection_type=ConnectionType.CLIENT)
        self._recv_events = {}
        self._recv_buffer = defaultdict(lambda: list())  # type: ignore
        self._default_duri = default_dataspace_uri
        self.ws = ws

        self.timeout = timeout
        self.client_info.endpoint_capabilities['MaxWebSocketMessagePayloadSize'] = MAXPAYLOADSIZE
        self.__recvtask = asyncio.create_task(self.__recv__())
        self.safeMode = False
        self.retryPause = 15

    #
    # client
    #

    async def send(self, body: ETPModel):
        if self.ws.closed:
            await self.connect()
        try:
            correlation_id = await self._send(body)
        except websockets.ConnectionClosed:
            logger.warning(f"Connection closed. Attempt to reconnection after {self.retryPause}s")
            time.sleep(self.retryPause)
            await self.connect()
            logger.warning("Reconnect successful")
            correlation_id = await self._send(body)
        return await self._recv(correlation_id)

        
    async def _send(self, body: ETPModel):

        msg = Message.get_object_message(
            body, message_flags=MessageFlags.FINALPART
        )
        if msg == None:
            raise TypeError(f"{type(body)} not valid etp protocol")

        msg.header.message_id = self.consume_msg_id()
        logger.debug(f"sending {msg.body.__class__.__name__} {repr(msg.header)}")

        # create future recv event
        self._recv_events[msg.header.message_id] = asyncio.Event()

        async for msg_part in msg.encode_message_generator(self.max_size, self):
            await self.ws.send(msg_part)

        return msg.header.message_id

    async def _recv(self, correlation_id: int) -> ETPModel:
        assert correlation_id in self._recv_events, "trying to recv response on non-existing message"

        try:
            async with timeout(self.timeout):
                await self._recv_events[correlation_id].wait()
        except asyncio.CancelledError as e:
            raise TimeoutError(f'Timeout before reciving {correlation_id=}') from e

        # cleanup
        bodies = self._clear_msg_on_buffer(correlation_id)

        for body in bodies:
            if isinstance(body, ProtocolException):
                logger.debug(body)
                raise ETPError.from_proto(body)

        if len(bodies) > 1:
            logger.warning(f"Recived {len(bodies)} messages, but only expected one")

        return bodies[0]

    async def close(self, reason=''):
        try:
            await self._send(CloseSession(reason=reason))
        finally:
            await self.ws.close(reason=reason)
            self.is_connected = False
            self.__recvtask.cancel("stopped")

            if len(self._recv_buffer):
                logger.error(f"Closed connection - but had stuff left in buffers ({len(self._recv_buffer)})")
                # logger.warning(self._recv_buffer)  # may contain data so lets not flood logs

    #
    #
    #

    def _clear_msg_on_buffer(self, correlation_id: int):
        del self._recv_events[correlation_id]
        return self._recv_buffer.pop(correlation_id)

    def _add_msg_to_buffer(self, msg: Message):
        self._recv_buffer[msg.header.correlation_id].append(msg.body)

        # NOTE: should we add task to autoclear buffer message if never waited on ?
        if msg.is_final_msg():
            self._recv_events[msg.header.correlation_id].set()  # set response on send event

    async def __recv__(self):

        logger.debug(f"starting recv loop")

        while (True):
            msg_data = await self.ws.recv()
            msg = Message.decode_binary_message(
                T.cast(bytes, msg_data), ETPClient.generic_transition_table
            )

            if msg is None:
                logger.error(f"Could not parse {msg_data}")
                continue

            logger.debug(f"recv {msg.body.__class__.__name__} {repr(msg.header)}")
            self._add_msg_to_buffer(msg)

    async def connect(self):
        msg = await self.send(
            RequestSession(
                applicationName=SETTINGS.application_name,
                applicationVersion=SETTINGS.application_version,
                clientInstanceId=uuid.uuid4(),  # type: ignore
                requestedProtocols=[
                    SupportedProtocol(protocol=p.value, protocolVersion=Version(major=1, minor=2), role='store')
                    for p in [CommunicationProtocol.DISCOVERY, CommunicationProtocol.STORE, CommunicationProtocol.DATA_ARRAY, CommunicationProtocol.DATASPACE]
                ],
                supportedDataObjects=[SupportedDataObject(qualifiedType="resqml20.*"), SupportedDataObject(qualifiedType="eml20.*")],
                currentDateTime=self.timestamp,
                earliestRetainedChangeTime=0,
                endpointCapabilities=dict(
                    MaxWebSocketMessagePayloadSize=DataValue(item=self.max_size)
                )
            )
        )
        assert msg and isinstance(msg, OpenSession)

        self.is_connected = True

        # ignore this endpoint
        _ = msg.endpoint_capabilities.pop('MessageQueueDepth', None)
        self.client_info.negotiate(msg)

        return self

    #

    @property
    def max_size(self):
        return self.client_info.getCapability("MaxWebSocketMessagePayloadSize")

    @property
    def max_array_size(self):
        return self.max_size - 512  # maxsize - 512 bytes for header and body

    @property
    def timestamp(self):
        return int(datetime.datetime.now(datetime.timezone.utc).timestamp())

    @property
    def default_dataspace_uri(self):
        return self._default_duri

    @default_dataspace_uri.setter
    def default_dataspace_uri(self, v: T.Union[DataspaceURI , str , None]):
        self._default_duri = None if v is None else DataspaceURI.from_any(v)

    def get_dataspace_or_default_uri(self, ds: T.Union[DataspaceURI , str , None]) -> DataspaceURI:
        """Returns default dataspace or user spefied one"""

        if ds is not None:
            return DataspaceURI.from_any(ds)

        if self._default_duri is None:
            raise ValueError("Could not get dataspace from userinput or default")

        return self._default_duri

    #
    # dataspace
    #

    async def put_dataspaces(self, *uris: T.Union[DataspaceURI , str]):
        from etptypes.energistics.etp.v12.protocol.dataspace.put_dataspaces import \
            PutDataspaces
        from etptypes.energistics.etp.v12.protocol.dataspace.put_dataspaces_response import \
            PutDataspacesResponse

        _uris = list(map(DataspaceURI.from_any, uris))

        time = self.timestamp
        response = await self.send(
            PutDataspaces(dataspaces={
                d.raw_uri: Dataspace(uri=d.raw_uri, storeCreated=time, storeLastWrite=time, path=d.dataspace)
                for d in _uris
            })
        )
        assert isinstance(response, PutDataspacesResponse), "Expected PutDataspacesResponse"

        assert len(response.success) == len(uris), f"expected {len(uris)} success's"

        return response.success

    async def put_dataspaces_no_raise(self, *uris: T.Union[DataspaceURI , str]):
        try:
            return await self.put_dataspaces(*uris)
        except ETPError:
            pass

    async def delete_dataspaces(self, *uris:T.Union[DataspaceURI , str]):
        from etptypes.energistics.etp.v12.protocol.dataspace.delete_dataspaces import \
            DeleteDataspaces
        from etptypes.energistics.etp.v12.protocol.dataspace.delete_dataspaces_response import \
            DeleteDataspacesResponse

        _uris = list(map(str, uris))

        response = await self.send(DeleteDataspaces(uris=dict(zip(_uris, _uris))))
        assert isinstance(response, DeleteDataspacesResponse), "Expected DeleteDataspacesResponse"
        return response.success

    #
    # data objects
    #

    async def get_data_objects(self, *uris: T.Union[DataObjectURI , str]):

        from etptypes.energistics.etp.v12.protocol.store.get_data_objects import \
            GetDataObjects
        from etptypes.energistics.etp.v12.protocol.store.get_data_objects_response import \
            GetDataObjectsResponse

        _uris = list(map(str, uris))

        msg = await self.send(
            GetDataObjects(uris=dict(zip(_uris, _uris)))
        )
        assert isinstance(msg, GetDataObjectsResponse), "Expected dataobjectsresponse"
        assert len(msg.data_objects) == len(_uris), "Here we assume that all three objects fit in a single record"

        return [msg.data_objects[u] for u in _uris]

    async def put_data_objects(self, *objs: DataObject):

        from etptypes.energistics.etp.v12.protocol.store.put_data_objects import \
            PutDataObjects
        from etptypes.energistics.etp.v12.protocol.store.put_data_objects_response import \
            PutDataObjectsResponse

        response = await self.send(
            PutDataObjects(dataObjects={p.resource.name: p for p in objs})
        )
        # logger.info(f"objects {response=:}")
        assert isinstance(response, PutDataObjectsResponse), "Expected PutDataObjectsResponse"
        # assert len(response.success) == len(objs)  # might be 0 if objects exists

        return response.success

    async def get_resqml_objects(self, *uris: T.Union[DataObjectURI , str]) -> T.List[ro.AbstractObject]:
        data_objects = await self.get_data_objects(*uris)
        return utils_xml.parse_resqml_objects(data_objects)

    async def put_resqml_objects(self, *objs: ro.AbstractObject, dataspace: T.Union[DataspaceURI ,str , None] = None):
        from etptypes.energistics.etp.v12.datatypes.object.resource import \
            Resource

        time = self.timestamp
        duri = self.get_dataspace_or_default_uri(dataspace)

        uris = [DataObjectURI.from_obj(duri, obj) for obj in objs]

        dobjs = [DataObject(
            format="xml",
            data=utils_xml.resqml_to_xml(obj),
            resource=Resource(
                uri=uri.raw_uri,
                name=obj.citation.title if obj.citation else obj.__class__.__name__,
                lastChanged=time,
                storeCreated=time,
                storeLastWrite=time,
                activeStatus="Inactive",  # type: ignore
                sourceCount=None,
                targetCount=None
            )
        ) for uri, obj in zip(uris, objs)]

        response = await self.put_data_objects(*dobjs)
        return uris

    async def delete_data_objects(self, *uris: T.Union[DataObjectURI , str], pruneContainedObjects=False):
        from etptypes.energistics.etp.v12.protocol.store.delete_data_objects import \
            DeleteDataObjects
        from etptypes.energistics.etp.v12.protocol.store.delete_data_objects_response import \
            DeleteDataObjectsResponse

        _uris = list(map(str, uris))

        response = await self.send(
            DeleteDataObjects(
                uris=dict(zip(_uris, _uris)),
                pruneContainedObjects=pruneContainedObjects
            )
        )
        # logger.info(f"delete objects {response=:}")
        assert isinstance(response, DeleteDataObjectsResponse), "Expected DeleteDataObjectsResponse"

        return response.deleted_uris

    #
    # xtgeo
    #
    @staticmethod
    def check_inside(x:float,y:float,patch:ro.Grid2dPatch):
        xori= patch.geometry.points.supporting_geometry.origin.coordinate1
        yori = patch.geometry.points.supporting_geometry.origin.coordinate2 
        xmax = xori + (patch.geometry.points.supporting_geometry.offset[0].spacing.value*patch.geometry.points.supporting_geometry.offset[0].spacing.count)
        ymax = yori + (patch.geometry.points.supporting_geometry.offset[1].spacing.value*patch.geometry.points.supporting_geometry.offset[1].spacing.count)
        if x < xori:
            return False
        if y < yori:
            return False
        if x > xmax:
            return False
        if y > ymax:
            return False  
        return True
    
    @staticmethod
    def find_closest_index(x,y,patch:ro.Grid2dPatch):
        x_ind = (x-patch.geometry.points.supporting_geometry.origin.coordinate1)/patch.geometry.points.supporting_geometry.offset[0].spacing.value
        y_ind = (y-patch.geometry.points.supporting_geometry.origin.coordinate2)/patch.geometry.points.supporting_geometry.offset[1].spacing.value
        return round(x_ind),round(y_ind)
    
    async def get_surface_value_x_y(self, epc_uri: T.Union[DataObjectURI , str] ,gri_uri: T.Union[DataObjectURI , str], x: T.Union[int,float], y: T.Union[int,float], method:T.Literal["bilinear","nearest"]):
        gri, = await self.get_resqml_objects(gri_uri) # parallelized using subarray
        xori= gri.grid2d_patch.geometry.points.supporting_geometry.origin.coordinate1
        yori = gri.grid2d_patch.geometry.points.supporting_geometry.origin.coordinate2 
        xinc = gri.grid2d_patch.geometry.points.supporting_geometry.offset[0].spacing.value
        yinc = gri.grid2d_patch.geometry.points.supporting_geometry.offset[1].spacing.value
        max_x_index_in_gri = gri.grid2d_patch.geometry.points.supporting_geometry.offset[0].spacing.count
        max_y_index_in_gri = gri.grid2d_patch.geometry.points.supporting_geometry.offset[1].spacing.count
        buffer = 4
        if not self.check_inside(x,y,gri.grid2d_patch):
            logger.info(f"Points not inside {x}:{y} {gri}")
            return
        uid=DataArrayIdentifier(
                    uri=str(epc_uri), pathInResource=gri.grid2d_patch.geometry.points.zvalues.values.path_in_hdf_file
                )
        if max_x_index_in_gri <= 10 or max_y_index_in_gri <= 10:
            surf = await self.get_xtgeo_surface(epc_uri,gri_uri)
            return surf.get_value_from_xy((x,y),sampling=method)
        
        x_ind,y_ind=self.find_closest_index(x,y,gri.grid2d_patch)
        if method == "nearest":
            arr = await self.get_subarray(uid,[x_ind,y_ind],[1,1])
            return arr[0][0]
        min_x_ind = max(x_ind-(buffer/2),0)
        min_y_ind = max(y_ind-(buffer/2),0)
        count_x = min(max_x_index_in_gri-min_x_ind,buffer)
        count_y = min(max_y_index_in_gri-min_y_ind,buffer)
        ## shift start index to left if not enough buffer on right
        if count_x < buffer:
            x_index_to_add = 3- count_x
            min_x_ind_new = max(0,min_x_ind-x_index_to_add)
            count_x=count_x + min_x_ind-min_x_ind_new+1
            min_x_ind = min_x_ind_new
        if count_y < buffer:
            y_index_to_add = 3- count_y
            min_y_ind_new = max(0,min_y_ind-y_index_to_add)
            count_y=count_y + min_y_ind-min_y_ind_new+1
            min_y_ind = min_y_ind_new
        arr = await self.get_subarray(uid,[min_x_ind,min_y_ind],[count_x,count_y])
        new_x_ori = xori+(min_x_ind*xinc)
        new_y_ori = yori+(min_y_ind*yinc) 
        regridded = xtgeo.RegularSurface(
            ncol=arr.shape[0],
            nrow=arr.shape[1],
            xori=new_x_ori,
            yori=new_y_ori,
            xinc=xinc,
            yinc=yinc,
            rotation=0.0,
            values=arr.flatten(),
        )
        return regridded.get_value_from_xy((x,y))
    
    async def get_xtgeo_surface(self, epc_uri: T.Union[DataObjectURI , str] ,gri_uri: T.Union[DataObjectURI , str]):
        gri, = await self.get_resqml_objects(gri_uri)

        # some checks
        
        assert isinstance(gri, ro.Grid2dRepresentation), "obj must be Grid2dRepresentation"
        sgeo = gri.grid2d_patch.geometry.points.supporting_geometry  # type: ignore
        if sys.version_info[1] != 10:
            assert isinstance(gri.grid2d_patch.geometry.points, ro.Point3dZValueArray), "Points must be Point3dZValueArray"
            assert isinstance(gri.grid2d_patch.geometry.points.zvalues, ro.DoubleHdf5Array), "Values must be DoubleHdf5Array"
            assert isinstance(gri.grid2d_patch.geometry.points.supporting_geometry, ro.Point3dLatticeArray), "Points support_geo must be Point3dLatticeArray"
            assert isinstance(sgeo, ro.Point3dLatticeArray), "supporting_geometry must be Point3dLatticeArray"
        assert isinstance(gri.grid2d_patch.geometry.points.zvalues.values, ro.Hdf5Dataset), "Values must be Hdf5Dataset"
        
        # get array
        array = await self.get_array(
            DataArrayIdentifier(
                uri=str(epc_uri), pathInResource=gri.grid2d_patch.geometry.points.zvalues.values.path_in_hdf_file
            )
        )

        return RegularSurface(
            ncol=array.shape[0], nrow=array.shape[1],
            xinc=sgeo.offset[0].spacing.value, yinc=sgeo.offset[1].spacing.value,  # type: ignore
            xori=sgeo.origin.coordinate1, yori=sgeo.origin.coordinate2,
            values=array,  # type: ignore
            masked=True
        )

    async def put_xtgeo_surface(self, surface: RegularSurface, epsg_code=23031, dataspace: T.Union[DataspaceURI , str , None] = None):
        """Returns (epc_uri, crs_uri, gri_uri)"""
        assert surface.values is not None, "cannot upload empty surface"

        epc, crs, gri = utils_xml.parse_xtgeo_surface_to_resqml_grid(surface, epsg_code)
        epc_uri, crs_uri, gri_uri = await self.put_resqml_objects(epc, crs, gri, dataspace=dataspace)
        response = await self.put_array(
            DataArrayIdentifier(
                uri=epc_uri.raw_uri if isinstance(epc_uri, DataObjectURI) else epc_uri,
                pathInResource=gri.grid2d_patch.geometry.points.zvalues.values.path_in_hdf_file  # type: ignore
            ),
            surface.values.filled(np.nan).astype(np.float32)
        )

        return epc_uri, crs_uri, gri_uri

    #
    # resqpy meshes
    #

    async def get_epc_mesh(self, epc_uri: T.Union[DataObjectURI , str], uns_uri: T.Union[DataObjectURI , str]):
        uns, = await self.get_resqml_objects(uns_uri)

        # some checks
        assert isinstance(uns, ro.UnstructuredGridRepresentation), "obj must be Grid2dRepresentation"
        assert isinstance(uns.geometry, ro.UnstructuredGridGeometry), "geometry must be UnstructuredGridGeometry"
        if sys.version_info[1] != 10:
            assert isinstance(uns.geometry.points, ro.Point3dHdf5Array), "points must be Point3dHdf5Array"
            assert isinstance(uns.geometry.faces_per_cell.elements, ro.IntegerHdf5Array), "faces_per_cell must be IntegerHdf5Array"
            assert isinstance(uns.geometry.faces_per_cell.cumulative_length, ro.IntegerHdf5Array), "faces_per_cell cl must be IntegerHdf5Array"
        assert isinstance(uns.geometry.points.coordinates, ro.Hdf5Dataset), "coordinates must be Hdf5Dataset"

        # # get array
        points = await self.get_array(
            DataArrayIdentifier(
                uri=str(epc_uri), pathInResource=uns.geometry.points.coordinates.path_in_hdf_file
            )
        )
        nodes_per_face = await self.get_array(
            DataArrayIdentifier(
                uri=str(epc_uri), pathInResource=uns.geometry.nodes_per_face.elements.values.path_in_hdf_file
            )
        )
        nodes_per_face_cl = await self.get_array(
            DataArrayIdentifier(
                uri=str(epc_uri), pathInResource=uns.geometry.nodes_per_face.cumulative_length.values.path_in_hdf_file
            )
        )
        faces_per_cell = await self.get_array(
            DataArrayIdentifier(
                uri=str(epc_uri), pathInResource=uns.geometry.faces_per_cell.elements.values.path_in_hdf_file
            )
        )
        faces_per_cell_cl = await self.get_array(
            DataArrayIdentifier(
                uri=str(epc_uri), pathInResource=uns.geometry.faces_per_cell.cumulative_length.values.path_in_hdf_file
            )
        )
        cell_face_is_right_handed = await self.get_array(
            DataArrayIdentifier(
                uri=str(epc_uri), pathInResource=uns.geometry.cell_face_is_right_handed.values.path_in_hdf_file
            )
        )

        return uns, points, nodes_per_face, nodes_per_face_cl, faces_per_cell, faces_per_cell_cl, cell_face_is_right_handed

    async def get_epc_mesh_property(self, epc_uri: T.Union[DataObjectURI , str], prop_uri: T.Union[DataObjectURI , str]):
        cprop0, = await self.get_resqml_objects(prop_uri)

        # some checks
        assert isinstance(cprop0, ro.ContinuousProperty) or isinstance(cprop0, ro.DiscreteProperty), "prop must be a Property"
        assert len(cprop0.patch_of_values) == 1, "property obj must have exactly one patch of values"

        # # get array
        values = await self.get_array(
            DataArrayIdentifier(
                uri=str(epc_uri), pathInResource=cprop0.patch_of_values[0].values.values.path_in_hdf_file,
            )
        )

        return cprop0, values
    
    @staticmethod
    def check_bound(points, x:float, y:float):
        if x > points[:,0].max() or x < points[:,0].min():
            return False
        if y > points[:,1].max() or y < points[:,1].min():
            return False
        return True
    
    async def get_epc_mesh_property_x_y(self, epc_uri: T.Union[DataObjectURI , str], uns_uri: T.Union[DataObjectURI , str], prop_uri: T.Union[DataObjectURI , str], x: float, y: float):
        uns, = await self.get_resqml_objects(uns_uri)
        points = await self.get_array(
                    DataArrayIdentifier(uri=str(epc_uri), pathInResource=uns.geometry.points.coordinates.path_in_hdf_file))
        chk = self.check_bound(points, x, y)
        if chk == False:
            return None
        unique_y=np.unique(points[:,1])
        y_smaller_sorted = np.sort(unique_y[np.argwhere(unique_y<y).flatten()])
        if y_smaller_sorted.size > 1:
            y_floor = y_smaller_sorted[-2]
        elif y_smaller_sorted.size == 1:
            y_floor = y_smaller_sorted[-1]
        else:
            pass
        y_larger_sorted  = np.sort(unique_y[np.argwhere(unique_y>y).flatten()])
        if y_larger_sorted.size > 1:
            y_ceil = y_larger_sorted[1]
        elif y_larger_sorted.size == 1:
            y_ceil=y_larger_sorted[0]
        else:
            pass
        start_new_row_idx = np.argwhere(np.diff(points[:,1]) !=0).flatten() +1

        to_fetch = []
        initial_result_arr_idx = 0
        for i in range(start_new_row_idx.size-1):
            sliced = points[start_new_row_idx[i]:start_new_row_idx[i+1],:]
            if sliced[0,1] <= y_ceil and sliced[0,1] >= y_floor:
                # Found slice that has same y
                x_diff = sliced[:,0]-x
                if all([np.any((x_diff>=0)), np.any((x_diff<=0))]): # y within this slice
                    first_idx = start_new_row_idx[i]
                    count = start_new_row_idx[i+1]-first_idx
                    to_fetch.append([start_new_row_idx[i],start_new_row_idx[i+1], count,initial_result_arr_idx])
                    initial_result_arr_idx+=count

        total_points_filtered = sum([i[2] for i in to_fetch])


        cprop, = await self.get_resqml_objects(prop_uri)
        assert str(cprop.indexable_element) == 'IndexableElements.NODES'
        props_uid = DataArrayIdentifier(
                    uri=str(epc_uri), pathInResource=cprop.patch_of_values[0].values.values.path_in_hdf_file)
        meta, = await self.get_array_metadata(props_uid)
        filtered_points = np.zeros((total_points_filtered,3), dtype=np.float64)
        all_values = np.empty(total_points_filtered, dtype=np.float64)
        async def populate(i):
            end_indx = i[2]+i[3]
            filtered_points[i[3]:end_indx] = points[i[0]:i[1]]
            if utils_arrays.get_nbytes(meta)* i[2]/points.shape[0] > self.max_array_size:
                all_values[i[3]:end_indx] = await self._get_array_chuncked(props_uid, i[0], i[2])
            else:
                all_values[i[3]:end_indx] = await self.get_subarray(props_uid, [i[0]], [i[2]])
            return
        
        r = await asyncio.gather(*[populate(i) for i in to_fetch], return_exceptions=True)
        if any(r):
            time.sleep(self.retryPause)
            if self.ws.closed:
                await self.connect()

            [await populate(to_fetch[idx]) for idx, res in enumerate(r) if isinstance(res, type(None)) == False]
        if isinstance(cprop, ro.DiscreteProperty):
            method = "nearest"
        else:
            method = "linear"
        
        # resolution= np.mean(np.diff(filtered[:,-1]))
        top = round(np.min(filtered_points[:,-1]),1)
        base = round(np.max(filtered_points[:,-1]),1)
        requested_depth = np.arange(top,base+1,100)
        requested_depth = requested_depth[requested_depth>0]
        request = np.tile([x, y, 0], (requested_depth.size,1))
        request[:,2]=requested_depth
        interpolated = griddata(filtered_points, all_values, request, method=method)
        response = np.vstack((requested_depth,interpolated))
        response_filtered = response[:, ~np.isnan(response[1])]
        return {"depth": response_filtered[0], "values": response_filtered[1]}
       

    async def put_epc_mesh(
        self, epc_filename, title_in, property_titles, projected_epsg, dataspace
    ):
        uns, crs, epc, timeseries, hexa = utils_xml.convert_epc_mesh_to_resqml_mesh(epc_filename, title_in, projected_epsg)
        epc_uri, crs_uri, uns_uri = await self.put_resqml_objects(epc, crs, uns, dataspace=dataspace)
        timeseries_uri = ""
        if timeseries is not None:
            timeseries_uris = await self.put_resqml_objects(timeseries, dataspace=dataspace)
            timeseries_uri = list(timeseries_uris)[0] if (len(list(timeseries_uris)) > 0) else ""
        
        # print("put_epc_mesh", uns, crs, epc, timeseries)
        # print("put_epc_mesh", epc_uri, crs_uri, uns_uri, timeseries_uri)
        print("put_epc_mesh property_titles", property_titles)

        #
        # mesh geometry (six arrays)
        #
        response = await self.put_array(
            DataArrayIdentifier(
                uri=epc_uri.raw_uri if isinstance(epc_uri, DataObjectURI) else epc_uri,
                pathInResource=uns.geometry.points.coordinates.path_in_hdf_file
            ),
            hexa.points_cached  # type: ignore
        )

        response = await self.put_array(
            DataArrayIdentifier(
                uri=epc_uri.raw_uri if isinstance(epc_uri, DataObjectURI) else epc_uri,
                pathInResource=uns.geometry.nodes_per_face.elements.values.path_in_hdf_file
            ),
            hexa.nodes_per_face.astype(np.int32)  # type: ignore
        )

        response = await self.put_array(
            DataArrayIdentifier(
                uri=epc_uri.raw_uri if isinstance(epc_uri, DataObjectURI) else epc_uri,
                pathInResource=uns.geometry.nodes_per_face.cumulative_length.values.path_in_hdf_file
            ),
            hexa.nodes_per_face_cl  # type: ignore
        )

        response = await self.put_array(
            DataArrayIdentifier(
                uri=epc_uri.raw_uri if isinstance(epc_uri, DataObjectURI) else epc_uri,
                pathInResource=uns.geometry.faces_per_cell.elements.values.path_in_hdf_file
            ),
            hexa.faces_per_cell  # type: ignore
        )

        response = await self.put_array(
            DataArrayIdentifier(
                uri=epc_uri.raw_uri if isinstance(epc_uri, DataObjectURI) else epc_uri,
                pathInResource=uns.geometry.faces_per_cell.cumulative_length.values.path_in_hdf_file
            ),
            hexa.faces_per_cell_cl  # type: ignore
        )

        response = await self.put_array(
            DataArrayIdentifier(
                uri=epc_uri.raw_uri if isinstance(epc_uri, DataObjectURI) else epc_uri,
                pathInResource=uns.geometry.cell_face_is_right_handed.values.path_in_hdf_file
            ),
            hexa.cell_face_is_right_handed  # type: ignore
        )

        #
        # mesh properties: one Property, one array of values, and an optional PropertyKind per property
        #
        prop_rddms_uris = {}
        for propname in property_titles:
            if timeseries is not None:
                time_indices = list(range(len(timeseries.time)))
                # print(f"prop {propname}: len of timeseries: {len(timeseries.time)}")
                print(propname)
                cprop0s, props, propertykind0 = utils_xml.convert_epc_mesh_property_to_resqml_mesh(epc_filename, hexa, propname, uns, epc, timeseries=timeseries, time_indices=time_indices)
            else:
                time_indices = [-1]
                cprop0s, props, propertykind0 = utils_xml.convert_epc_mesh_property_to_resqml_mesh(epc_filename, hexa, propname, uns, epc)
            if isinstance(cprop0s, type(None)):
                continue
            cprop_uris = []
            for cprop0, prop, time_index in zip(cprop0s, props, time_indices):
                assert isinstance(cprop0, ro.ContinuousProperty) or isinstance(cprop0, ro.DiscreteProperty), "prop must be a Property"
                assert len(cprop0.patch_of_values) == 1, "property obj must have exactly one patch of values"

                propkind_uri = [""] if (propertykind0 is None) else (await self.put_resqml_objects(propertykind0, dataspace=dataspace))
                cprop_uri = await self.put_resqml_objects(cprop0, dataspace=dataspace)
                
                response = await self.put_array(
                    DataArrayIdentifier(
                        uri=epc_uri.raw_uri if isinstance(epc_uri, DataObjectURI) else epc_uri,
                        pathInResource=cprop0.patch_of_values[0].values.values.path_in_hdf_file,
                    ),
                    prop.array_ref(),  # type: ignore
                )
                cprop_uris.extend(cprop_uri)
            prop_rddms_uris[propname] = [propkind_uri, cprop_uris]

        return [epc_uri, crs_uri, uns_uri, timeseries_uri], prop_rddms_uris
    async def get_mesh_points(self, epc_uri:T.Union[DataObjectURI , str], uns_uri:T.Union[DataObjectURI , str]):
        uns, = await self.get_resqml_objects(uns_uri)
        points = await self.get_array(
                DataArrayIdentifier(
                    uri=str(epc_uri), pathInResource=uns.geometry.points.coordinates.path_in_hdf_file
                )
            )
        return points
    
    async def get_epc_property_surface_slice_node(self,epc_uri:T.Union[DataObjectURI , str], cprop0: ro.AbstractObject, points: np.ndarray, node_index: int, n_node_per_pos: int):
        #indexing_array = np.arange(0, points.shape[0], 1, dtype=np.int32)[node_index::n_node_per_pos]
        indexing_array = np.arange(node_index, points.shape[0], n_node_per_pos, dtype=np.int32)
        results = points[indexing_array,:]
        arr = await asyncio.gather(*[self.get_subarray( DataArrayIdentifier(
                uri=str(epc_uri), pathInResource=cprop0.patch_of_values[0].values.values.path_in_hdf_file,),
                [i], [1]) for i in indexing_array])
        arr =np.array(arr).flatten()
        assert results.shape[0] == arr.size
        results[:,2] = arr
        return results
    
    async def get_epc_property_surface_slice_cell(self,epc_uri:T.Union[DataObjectURI , str], cprop0: ro.AbstractObject, points: np.ndarray, node_index: int, n_node_per_pos: int, get_cell_pos=True):
        m,=await self.get_array_metadata(DataArrayIdentifier(
                    uri=str(epc_uri), pathInResource=cprop0.patch_of_values[0].values.values.path_in_hdf_file,))
        n_cells = m.dimensions[0]
        layers_per_sediment_unit=2
        n_cell_per_pos = n_node_per_pos -1
        indexing_array = np.arange(node_index,n_cells,n_cell_per_pos, dtype=np.int32)
        if get_cell_pos:
            results = utils_arrays.get_cells_positions(points, n_cells, n_cell_per_pos, layers_per_sediment_unit, n_node_per_pos, node_index)
        else:
            results = np.zeros((int(n_cells/n_cell_per_pos),3), dtype=np.float64)
        arr = await asyncio.gather(*[self.get_subarray( DataArrayIdentifier(
                uri=str(epc_uri), pathInResource=cprop0.patch_of_values[0].values.values.path_in_hdf_file,),
                [i], [1]) for i in indexing_array])
        arr =np.array(arr).flatten()
        assert results.shape[0] == arr.size
        results[:,2] = arr
        return results


    async def get_epc_property_surface_slice(self, epc_uri:T.Union[DataObjectURI , str], uns_uri:T.Union[DataObjectURI , str], prop_uri:T.Union[DataObjectURI , str], node_index: int, n_node_per_pos: int):
        # n_node_per_pos number of nodes in a 1D location
        # node_index index of slice from top. Warmth has 2 nodes per sediment layer. E.g. top of second layer will have index 2
        points = await self.get_mesh_points(epc_uri, uns_uri)
        cprop0, = await self.get_resqml_objects(prop_uri)
        prop_at_node = False
        if str(cprop0.indexable_element) == 'IndexableElements.NODES':
            prop_at_node = True
        # node_per_sed = 2
        # n_sed_node = n_sed *node_per_sed
        # n_crust_node = 4
        # n_node_per_pos = n_sed_node + n_crust_node
        # start_idx_pos = sediment_id *node_per_sed
        if prop_at_node:
            return await self.get_epc_property_surface_slice_node(epc_uri, cprop0, points,node_index, n_node_per_pos)
        else:
            return await self.get_epc_property_surface_slice_cell(epc_uri, cprop0, points,node_index, n_node_per_pos)
    
    async def get_epc_property_surface_slice_xtgeo(self, epc_uri:T.Union[DataObjectURI , str], uns_uri:T.Union[DataObjectURI , str], prop_uri:T.Union[DataObjectURI , str], node_index: int, n_node_per_pos: int):
        data = await self.get_epc_property_surface_slice(epc_uri, uns_uri,prop_uri,node_index, n_node_per_pos)
        return utils_arrays.grid_xtgeo(data)
    
    #
    # array
    #

    async def get_array_metadata(self, *uids: DataArrayIdentifier):
        from etptypes.energistics.etp.v12.protocol.data_array.get_data_array_metadata import \
            GetDataArrayMetadata
        from etptypes.energistics.etp.v12.protocol.data_array.get_data_array_metadata_response import \
            GetDataArrayMetadataResponse

        response = await self.send(
            GetDataArrayMetadata(dataArrays={i.path_in_resource: i for i in uids})
        )
        assert isinstance(response, GetDataArrayMetadataResponse)

        if len(response.array_metadata) != len(uids):
            raise ETPError(f'Not all uids found ({uids})', 11)

        # return in same order as arguments
        return [response.array_metadata[i.path_in_resource] for i in uids]

    async def get_array(self, uid: DataArrayIdentifier):
        from etptypes.energistics.etp.v12.protocol.data_array.get_data_arrays import \
            GetDataArrays
        from etptypes.energistics.etp.v12.protocol.data_array.get_data_arrays_response import \
            GetDataArraysResponse

        # Check if we can upload the full array in one go.
        meta, = await self.get_array_metadata(uid)
        if utils_arrays.get_nbytes(meta) > self.max_array_size:
            return await self._get_array_chuncked(uid)

        response = await self.send(
            GetDataArrays(dataArrays={uid.path_in_resource: uid})
        )
        assert isinstance(response, GetDataArraysResponse), "Expected GetDataArraysResponse"

        arrays = list(response.data_arrays.values())
        return utils_arrays.to_numpy(arrays[0])

    async def put_array(self, uid: DataArrayIdentifier, data: np.ndarray):
        from etptypes.energistics.etp.v12.datatypes.data_array_types.put_data_arrays_type import \
            PutDataArraysType
        from etptypes.energistics.etp.v12.protocol.data_array.put_data_arrays import \
            PutDataArrays
        from etptypes.energistics.etp.v12.protocol.data_array.put_data_arrays_response import \
            PutDataArraysResponse

        # Check if we can upload the full array in one go.
        if data.nbytes > self.max_array_size:
            return await self._put_array_chuncked(uid, data)

        response = await self.send(
            PutDataArrays(
                dataArrays={uid.path_in_resource: PutDataArraysType(uid=uid, array=utils_arrays.to_data_array(data))})
        )
        assert isinstance(response, PutDataArraysResponse), "Expected PutDataArraysResponse"
        assert len(response.success) == 1, "expected one success from put_array"
        return response.success

    async def get_subarray(self, uid: DataArrayIdentifier, starts: T.Union[np.ndarray , T.List[int]], counts: T.Union[np.ndarray , T.List[int]]):
        starts = np.array(starts).astype(np.int64)
        counts = np.array(counts).astype(np.int64)

        from etptypes.energistics.etp.v12.datatypes.data_array_types.get_data_subarrays_type import \
            GetDataSubarraysType
        from etptypes.energistics.etp.v12.protocol.data_array.get_data_subarrays import \
            GetDataSubarrays
        from etptypes.energistics.etp.v12.protocol.data_array.get_data_subarrays_response import \
            GetDataSubarraysResponse

        logger.debug(f"get_subarray {starts=:} {counts=:}")

        payload = GetDataSubarraysType(
            uid=uid,
            starts=starts.tolist(),
            counts=counts.tolist(),
        )
        response = await self.send(
            GetDataSubarrays(dataSubarrays={uid.path_in_resource: payload})
        )
        assert isinstance(response, GetDataSubarraysResponse), "Expected GetDataSubarraysResponse"

        arrays = list(response.data_subarrays.values())
        return utils_arrays.to_numpy(arrays[0])

    async def put_subarray(self, uid: DataArrayIdentifier, data: np.ndarray, starts: T.Union[np.ndarray ,T.List[int]], counts: T.Union[np.ndarray , T.List[int]], put_uninitialized=False):
        from etptypes.energistics.etp.v12.datatypes.data_array_types.put_data_subarrays_type import \
            PutDataSubarraysType
        from etptypes.energistics.etp.v12.protocol.data_array.put_data_subarrays import \
            PutDataSubarrays
        from etptypes.energistics.etp.v12.protocol.data_array.put_data_subarrays_response import \
            PutDataSubarraysResponse
        # starts [start_X, starts_Y]
        # counts [count_X, count_Y]
        starts = np.array(starts).astype(np.int64) # len = 2 [x_start_index, y_start_index]
        counts = np.array(counts).astype(np.int64) # len = 2
        ends = starts + counts # len = 2
        if put_uninitialized:
            transport_array_type = utils_arrays.get_transport(data.dtype)
            await self._put_uninitialized_data_array(uid, data.shape, transport_array_type=transport_array_type)

        slices = tuple(map(lambda se: slice(se[0], se[1]), zip(starts, ends)))
        dataarray = utils_arrays.to_data_array(data[slices])
        payload = PutDataSubarraysType(
            uid=uid,
            data=dataarray.data,
            starts=starts.tolist(),
            counts=counts.tolist(),
        )

        logger.debug(f"put_subarray {data.shape=:} {starts=:} {counts=:} {dataarray.data.item.__class__.__name__}")

        response = await self.send(
            PutDataSubarrays(dataSubarrays={uid.path_in_resource: payload})
        )
        assert isinstance(response, PutDataSubarraysResponse), "Expected PutDataSubarraysResponse"

        assert len(response.success) == 1, "expected one success"
        return response.success

    #
    # chuncked get array - ETP will not chunck response - so we need to do it manually
    #

    def _get_chunk_sizes(self, shape, dtype: np.dtype[T.Any] = np.dtype(np.float32), offset = 0):
        shape = np.array(shape)

        # capsize blocksize
        max_items = self.max_array_size / dtype.itemsize  # remove 512 bytes for headers and body
        block_size = np.power(max_items, 1. / len(shape))
        block_size = min(2048, int(block_size // 2) * 2)

        assert block_size > 8, "computed blocksize unreasonable small"

        all_ranges = [range(s // block_size + 1) for s in shape]
        indexes = np.array(np.meshgrid(*all_ranges)).T.reshape(-1, len(shape))

        for ijk in indexes:
            starts = ijk * block_size
            if offset != 0:
                starts = starts + offset
            ends = np.fmin(shape, starts + block_size)
            if offset != 0:
                ends = ends + offset
            counts = ends - starts
            if any(counts == 0):
                continue
            yield starts, counts

    async def _get_array_chuncked(self, uid: DataArrayIdentifier, offset:int = 0, total_count: T.Union[int, None] = None):

        metadata = (await self.get_array_metadata(uid))[0]
        if len(metadata.dimensions) != 1 and offset != 0:
            raise Exception("Offset is only implemented for 1D array")

        if isinstance(total_count, (int, float)):
            buffer_shape = np.array([total_count], dtype=np.int64)
        else:
            buffer_shape = np.array(metadata.dimensions, dtype=np.int64)

        dtype = utils_arrays.get_dtype(metadata.transport_array_type)
        buffer = np.zeros(buffer_shape, dtype=dtype)
        params = []
        async def populate(starts, counts):
            params.append([starts, counts])
            array = await self.get_subarray(uid, starts, counts)
            ends = starts + counts
            slices = tuple(map(lambda se: slice(se[0], se[1]), zip(starts-offset, ends-offset)))
            buffer[slices] = array
            return

        r = await asyncio.gather(*[
            populate(starts, counts)
            for starts, counts in self._get_chunk_sizes(buffer_shape, dtype, offset)
        ], return_exceptions=True)
        if any(r):
            time.sleep(self.retryPause)
            if self.ws.closed:
                await self.connect()

            [await populate(params[idx][0], params[idx][1]) for idx, res in enumerate(r) if isinstance(res, type(None)) == False]
        return buffer

    async def _put_array_chuncked(self, uid: DataArrayIdentifier, data: np.ndarray):

        transport_array_type = utils_arrays.get_transport(data.dtype)
        await self._put_uninitialized_data_array(uid, data.shape, transport_array_type=transport_array_type)
        params = []
        coro = []
        for starts, counts in self._get_chunk_sizes(data.shape, data.dtype):
            params.append([starts, counts])
            coro.append(self.put_subarray(uid, data, starts, counts))
        r = await asyncio.gather(*coro, return_exceptions=True)
        if any(r):
            time.sleep(self.retryPause)
            if self.ws.closed:
                await self.connect()
            [await self.put_subarray(uid, data, params[idx][0], params[idx][1]) for idx, res in enumerate(r) if isinstance(res, type(None)) == False]
        return {uid.uri: ''}

    async def _put_uninitialized_data_array(self, uid: DataArrayIdentifier, shape: T.Tuple[int, ...], transport_array_type=AnyArrayType.ARRAY_OF_FLOAT, logical_array_type=AnyLogicalArrayType.ARRAY_OF_BOOLEAN):
        from etptypes.energistics.etp.v12.datatypes.data_array_types.put_uninitialized_data_array_type import \
            PutUninitializedDataArrayType
        from etptypes.energistics.etp.v12.protocol.data_array.put_uninitialized_data_arrays import \
            PutUninitializedDataArrays
        from etptypes.energistics.etp.v12.protocol.data_array.put_uninitialized_data_arrays_response import \
            PutUninitializedDataArraysResponse

        payload = PutUninitializedDataArrayType(
            uid=uid,
            metadata=(DataArrayMetadata(
                dimensions=list(shape),  # type: ignore
                transportArrayType=transport_array_type,
                logicalArrayType=logical_array_type,
                storeLastWrite=self.timestamp,
                storeCreated=self.timestamp,
            ))
        )
        response = await self.send(
            PutUninitializedDataArrays(dataArrays={uid.path_in_resource: payload})
        )
        assert isinstance(response, PutUninitializedDataArraysResponse), "Expected PutUninitializedDataArraysResponse"
        assert len(response.success) == 1, "expected one success"
        return response.success


# define an asynchronous context manager
class connect:

    def __init__(self, authorization: T.Optional[SecretStr] = None, timeout=10.):
        self.server_url = SETTINGS.etp_url
        self.authorization = authorization
        self.timeout = timeout
        self.default_dataspace_uri = DataspaceURI.from_name(SETTINGS.dataspace)

    # enter the async context manager
    async def __aenter__(self):
        if isinstance(self.authorization, str):
            token = self.authorization
        elif isinstance(self.authorization, SecretStr):
            token = self.authorization.get_secret_value()
        ws = await websockets.connect(
            self.server_url,
            subprotocols=[ETPClient.SUB_PROTOCOL],  # type: ignore
            extra_headers={"Authorization": token} if self.authorization else {},
            max_size=MAXPAYLOADSIZE,
            ping_timeout=self.timeout,
            open_timeout=None,
        )
        self.client = ETPClient(ws, default_dataspace_uri=self.default_dataspace_uri, timeout=self.timeout)
        await self.client.connect()

        return self.client

    # exit the async context manager
    async def __aexit__(self, exc_type, exc: Exception, tb: TracebackType):
        await self.client.close()




import typing as T
import numpy as np
from scipy import interpolate
import xtgeo

from .types import (AnyArray, AnyArrayType, ArrayOfBoolean, ArrayOfDouble,
                    ArrayOfFloat, ArrayOfInt, ArrayOfLong, DataArray,
                    DataArrayMetadata)

SUPPORED_ARRAY_TYPES = T.Union[ArrayOfFloat , ArrayOfBoolean, ArrayOfInt, ArrayOfLong , ArrayOfDouble]

_ARRAY_MAP_TYPES: dict[AnyArrayType, np.dtype[T.Any]] = {
    AnyArrayType.ARRAY_OF_FLOAT: np.dtype(np.float32),
    AnyArrayType.ARRAY_OF_DOUBLE: np.dtype(np.float64),
    AnyArrayType.ARRAY_OF_INT: np.dtype(np.int32),
    AnyArrayType.ARRAY_OF_LONG: np.dtype(np.int64),
    AnyArrayType.ARRAY_OF_BOOLEAN: np.dtype(np.bool_)
}

_ARRAY_MAP: dict[AnyArrayType, T.Type[SUPPORED_ARRAY_TYPES]] = {
    AnyArrayType.ARRAY_OF_FLOAT: ArrayOfFloat,
    AnyArrayType.ARRAY_OF_DOUBLE: ArrayOfDouble,
    AnyArrayType.ARRAY_OF_INT: ArrayOfInt,
    AnyArrayType.ARRAY_OF_LONG: ArrayOfLong,
    AnyArrayType.ARRAY_OF_BOOLEAN: ArrayOfBoolean
}


def get_transport_from_name(k: str):
    return AnyArrayType(k[0].lower() + k[1:])


def get_transport(dtype: np.dtype):

    arraytype = [item[0] for item in _ARRAY_MAP_TYPES.items() if item[1] == dtype]
    if not len(arraytype):
        raise TypeError(f"Not {type(dtype)} supported")

    return arraytype[0]


def get_cls(dtype: np.dtype):
    return _ARRAY_MAP[get_transport(dtype)]


def get_dtype(item: T.Union[AnyArray ,AnyArrayType]):
    atype = item if isinstance(item, AnyArrayType) else get_transport_from_name(item.item.__class__.__name__)

    if atype not in _ARRAY_MAP_TYPES:
        raise TypeError(f"Not {atype} supported")

    return _ARRAY_MAP_TYPES[atype]


def get_nbytes(md: DataArrayMetadata):
    dtype = get_dtype(md.transport_array_type)
    return int(np.prod(np.array(md.dimensions)) * dtype.itemsize)


def to_numpy(data_array: DataArray):
    dims: T.Tuple[int, ...] = tuple(map(int, data_array.dimensions))
    return np.asarray(
        data_array.data.item.values,  # type: ignore
        dtype=get_dtype(data_array.data)
    ).reshape(dims)


def to_data_array(data: np.ndarray):
    cls = get_cls(data.dtype)
    return DataArray(
        dimensions=data.shape,  # type: ignore
        data=AnyArray(item=cls(values=data.flatten().tolist()))
    )

def mid_point_rectangle(arr: np.ndarray):
    all_x=arr[:,0]
    all_y= arr[:,1]
    min_x = np.min(all_x)
    min_y = np.min(all_y)
    mid_x = ((np.max(all_x)-min_x)/2)+min_x
    mid_y = ((np.max(all_y)-min_y)/2)+min_y
    return np.array([mid_x, mid_y])

def grid_xtgeo(data: np.ndarray):
    max_x = np.nanmax(data[:,0])
    max_y = np.nanmax(data[:,1])
    min_x = np.nanmin(data[:,0])
    min_y = np.nanmin(data[:,1])
    u_x = np.sort(np.unique(data[:,0]))
    u_y = np.sort(np.unique(data[:,1]))
    xinc = u_x[1]- u_x[0]
    yinc = u_y[1]- u_y[0]
    grid_x, grid_y = np.mgrid[
        min_x: max_x + xinc: xinc,
        min_y: max_y + yinc: yinc,
    ]

    interp = interpolate.LinearNDInterpolator(data[:,:-1], data[:,-1], fill_value=np.nan, rescale=False)
    z = interp(np.array([grid_x.flatten(), grid_y.flatten() ]).T )
    zz = np.reshape(z,grid_x.shape)

    surf = xtgeo.RegularSurface(
        ncol=grid_x.shape[0],
        nrow=grid_x.shape[1],
        xori=min_x,
        yori=min_y,
        xinc=xinc,
        yinc=yinc,
        rotation=0.0,
        values=zz,
    )
    return surf

def get_cells_positions(points: np.ndarray, n_cells:int, n_cell_per_pos:int,layers_per_sediment_unit:int,n_node_per_pos:int,node_index: int):
    results = np.zeros((int(n_cells/n_cell_per_pos),3), dtype=np.float64)
    grid_x_pos = np.unique(points[:,0])
    grid_y_pos = np.unique(points[:,1])
    counter = 0
    # find cell index and location

    for y_ind in range(0,len(grid_y_pos)-1):
        for x_ind in range(0,len(grid_x_pos)-1):
            top_depth= []
            for corner_x in range(layers_per_sediment_unit):
                for corner_y in range(layers_per_sediment_unit):
                    node_indx = (( (y_ind+corner_y)*len(grid_x_pos) + (x_ind+corner_x) ) * n_node_per_pos)+ node_index
                    top_depth.append( points[node_indx])
            results[counter,0:2] = mid_point_rectangle(np.array(top_depth))
            counter+=1
    return results


import datetime
import typing as T
from uuid import uuid4

import lxml.etree as ET
import numpy as np
import xtgeo
from xsdata.formats.dataclass.context import XmlContext
from xsdata.formats.dataclass.parsers import XmlParser
from xsdata.formats.dataclass.serializers import XmlSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig
from xsdata.models.datatype import XmlDateTime

import pyetp.resqml_objects as ro
from pyetp.config import SETTINGS
import logging
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
from .types import DataObject

schema_version = "2.0"


def get_data_object_type(obj: ro.AbstractObject):
    return obj.__class__.__name__


def parse_resqml_objects(data_objects: T.List[DataObject]):
    # This function creates a list of resqml-objects from the returned xml from
    # the ETP-server. It dynamically finds the relevant resqml dataclass using
    # the object name found in the xml. Its intention is to be used after
    # calling the get_data_objects-protocol.

    # Set up an XML-parser from xsdata.
    parser = XmlParser(context=XmlContext())

    return [
        parser.from_bytes(
            data_object.data,
            getattr(ro, ET.QName(ET.fromstring(data_object.data).tag).localname),
        )
        for data_object in data_objects
    ]


def resqml_to_xml(obj: ro.AbstractObject):
    serializer = XmlSerializer(config=SerializerConfig())
    return str.encode(serializer.render(obj))


def create_common_citation(title: str):

    return ro.Citation(
        title=title,
        creation=XmlDateTime.from_string(
            datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S%z")
        ),
        originator=SETTINGS.application_name,
        format=f"{SETTINGS.application_name}:v{SETTINGS.application_version}",
    )


def create_common_crs(title: str, projected_epsg, rotation: float = 0.0):
    return ro.LocalDepth3dCrs(
        citation=create_common_citation(f"CRS for {title}"),
        schema_version=schema_version,
        uuid=str(uuid4()),
        # NOTE: I assume that we let the CRS have no offset, and add any offset
        # in the grid instead.
        xoffset=0.0,
        yoffset=0.0,
        zoffset=0.0,
        areal_rotation=ro.PlaneAngleMeasure(
            # Here rotation should be zero!
            value=rotation,
            uom=ro.PlaneAngleUom.DEGA,
        ),
        # NOTE: Verify that this is the projected axis order
        projected_axis_order=ro.AxisOrder2d.EASTING_NORTHING,
        projected_uom=ro.LengthUom.M,
        vertical_uom=ro.LengthUom.M,
        zincreasing_downward=True,
        vertical_crs=ro.VerticalUnknownCrs(
            unknown="unknown",
        ),
        projected_crs=ro.ProjectedCrsEpsgCode(
            epsg_code=projected_epsg,
        ),
    )


def create_epc(schema_version="2.0"):
    return ro.EpcExternalPartReference(
        citation=create_common_citation("Hdf Proxy"),
        schema_version=schema_version,
        uuid=str(uuid4()),
        mime_type="application/x-hdf5",
    )


def parse_xtgeo_surface_to_resqml_grid(surf: 'xtgeo.RegularSurface', projected_epsg: int):
    # Build the RESQML-objects "manually" from the generated dataclasses.
    # Their content is described also in the RESQML v2.0.1 standard that is
    # available for download here:
    # https://publications.opengroup.org/standards/energistics-standards/v231a

    title = surf.name or "regularsurface"

    # NOTE: xtgeo uses nrow for axis 1 in the array, and ncol for axis 0.  This
    # means that surf.nrow is the fastest changing axis, and surf.ncol the
    # slowest changing axis, and we have surf.values.shape == (surf.ncol,
    # surf.nrow). The author of this note finds that confusing, but such is
    # life.
    epc, crs, gri = instantiate_resqml_grid(title, surf.get_rotation(), surf.xori, surf.yori, surf.xinc, surf.yinc, surf.ncol, surf.nrow, projected_epsg)
    return epc, crs, gri

def instantiate_resqml_grid(name:str, rotation: float, x0: float, y0: float, dx: float, dy: float, nx: int, ny: int, epsg: int):
    assert np.abs(rotation) < 1e-7, "Maps should have no rotation!"

    epc = create_epc()
    crs = create_common_crs(name, epsg, rotation)   # Here rotation should be zero!


    gri = ro.Grid2dRepresentation(
        uuid=(grid_uuid := str(uuid4())),
        schema_version=schema_version,
        surface_role=ro.SurfaceRole.MAP,
        citation=create_common_citation(name),
        grid2d_patch=ro.Grid2dPatch(
            # TODO: Perhaps we can use this for tiling?
            patch_index=0,
            # NumPy-arrays are C-ordered, meaning that the last index is
            # the index that changes most rapidly. However, xtgeo uses nrow for
            # axis 1 in the array, and ncol for axis 0. This means that
            # surf.nrow is the fastest changing axis, and surf.ncol the slowest
            # changing axis (as surf.values.shape == (surf.ncol, surf.nrow))
            fastest_axis_count=ny,
            slowest_axis_count=nx,
            geometry=ro.PointGeometry(
                local_crs=ro.DataObjectReference(
                    # NOTE: See Energistics Identifier Specification 4.0
                    # (it is downloaded alongside the RESQML v2.0.1
                    # standard) section 4.1 for an explanation on the
                    # format of content_type.
                    content_type=f"application/x-resqml+xml;version={schema_version};type={get_data_object_type(crs)}",
                    title=crs.citation.title,
                    uuid=crs.uuid,
                ),
                points=ro.Point3dZValueArray(
                    supporting_geometry=ro.Point3dLatticeArray(
                        origin=ro.Point3d(
                            coordinate1=x0,
                            coordinate2=y0,
                            coordinate3=0.0,
                        ),
                        # NOTE: The ordering in the offset-list should be
                        # preserved when the data is passed back and forth.
                        # However, _we_ need to ensure a consistent ordering
                        # for ourselves. In this setup I have set the slowest
                        # axis to come first, i.e., the x-axis or axis 0 in
                        # NumPy. The reason is so that it corresponds with the
                        # origin above where "coordinate1" is set to be the
                        # x0-coordinate, and "coordinate2" the y0-coordinate.
                        # However, we can change this as we see fit.
                        offset=[
                            # Offset for x-direction, i.e., the slowest axis
                            ro.Point3dOffset(
                                offset=ro.Point3d(
                                    coordinate1=1.0,
                                    coordinate2=0.0,
                                    coordinate3=0.0,
                                ),
                                spacing=ro.DoubleConstantArray(
                                    value=dx,
                                    count=nx - 1,
                                ),
                            ),
                            # Offset for y-direction, i.e., the fastest axis
                            ro.Point3dOffset(
                                offset=ro.Point3d(
                                    coordinate1=0.0,
                                    coordinate2=1.0,
                                    coordinate3=0.0,
                                ),
                                spacing=ro.DoubleConstantArray(
                                    value=dy,
                                    count=ny - 1,
                                ),
                            ),
                        ],
                    ),
                    zvalues=ro.DoubleHdf5Array(
                        values=ro.Hdf5Dataset(
                            path_in_hdf_file=f"/RESQML/{grid_uuid}/zvalues",
                            hdf_proxy=ro.DataObjectReference(
                                content_type=f"application/x-eml+xml;version={schema_version};type={get_data_object_type(epc)}",
                                title=epc.citation.title,
                                uuid=epc.uuid,
                            ),
                        ),
                    ),
                ),
            ),
        ),
    )
    return epc, crs, gri

def convert_epc_mesh_to_resqml_mesh(epc_filename, title_in, projected_epsg):
    import numpy as np
    import resqpy.model as rq
    import resqpy.unstructured as rug
    import resqpy.property as rqp
    import resqpy.time_series as rts

    title = title_in or "hexamesh"

    model = rq.Model(epc_filename)
    assert model is not None

    #
    # read mesh:  vertex positions and cell definitions
    #
    hexa_uuid = model.uuid(obj_type='UnstructuredGridRepresentation', title=title_in)
    assert hexa_uuid is not None
    hexa = rug.HexaGrid(model, uuid=hexa_uuid)
    assert hexa is not None
    assert hexa.cell_shape == 'hexahedral'
    hexa.check_hexahedral()

    ts_uuid = model.uuid(obj_type = "TimeSeries")
    # ts_uuid_2 = model.uuid(obj_type='GeologicTimeSeries')
    # print("TS UUIDs: ", ts_uuid, ts_uuid_2)
    gts = rts.GeologicTimeSeries(model, uuid=ts_uuid)

    print("gts: ", gts)
    timeseries = None
    # dynamic_points = []    
    if (ts_uuid is not None) and (gts is not None):



        ro_timestamps = []
        for i in gts.iter_timestamps(as_string=False):
            ro_timestamps.append(
                ro.Timestamp(
                    date_time=XmlDateTime.from_string("0001-01-01T00:00:00.00+00:00"),
                    year_offset=int(i),
                )
            )

        print(f"Generating time series with {len(ro_timestamps)} indices, year offsets: {ro_timestamps[0].year_offset} -- {ro_timestamps[-1].year_offset}.")

        timeseries = ro.TimeSeries(
            citation=create_common_citation(str(gts.citation_title)),
            schema_version=schema_version,
            uuid=str(gts.uuid),
            time = ro_timestamps,
        )



    crs = create_common_crs(title, projected_epsg)

    epc = ro.EpcExternalPartReference(
        citation=create_common_citation("Hdf Proxy"),
        schema_version=schema_version,
        uuid=str(uuid4()),
        mime_type="application/x-hdf5",
    )

    cellshape = ro.CellShape.HEXAHEDRAL if (hexa.cell_shape == "hexahedral") else ro.CellShape.TETRAHEDRAL

    geom = ro.UnstructuredGridGeometry(
        local_crs=ro.DataObjectReference(
            content_type=f"application/x-resqml+xml;version={schema_version};type={get_data_object_type(crs)}",
            title=crs.citation.title,
            uuid=crs.uuid,
        ),
        node_count=hexa.node_count or -1,
        face_count=hexa.face_count or -1,
        cell_shape=cellshape,
        points=ro.Point3dHdf5Array(
            coordinates=ro.Hdf5Dataset(
                path_in_hdf_file=f"/RESQML/{str(hexa_uuid)}/points",
                hdf_proxy=ro.DataObjectReference(
                    content_type=f"application/x-eml+xml;version={schema_version};type={get_data_object_type(epc)}",
                    title=epc.citation.title,
                    uuid=str(epc.uuid),
                ),
            )
        ),
        nodes_per_face=ro.ResqmlJaggedArray(
            elements=ro.IntegerHdf5Array(
                null_value=-1,
                values=ro.Hdf5Dataset(
                    path_in_hdf_file=f"/RESQML/{str(hexa_uuid)}/nodes_per_face",
                    hdf_proxy=ro.DataObjectReference(
                        content_type=f"application/x-eml+xml;version={schema_version};type={get_data_object_type(epc)}",
                        title=epc.citation.title,
                        uuid=str(epc.uuid),
                    ),
                )
            ),
            cumulative_length=ro.IntegerHdf5Array(
                null_value=-1,
                values=ro.Hdf5Dataset(
                    path_in_hdf_file=f"/RESQML/{str(hexa_uuid)}/nodes_per_face_cl",
                    hdf_proxy=ro.DataObjectReference(
                        content_type=f"application/x-eml+xml;version={schema_version};type={get_data_object_type(epc)}",
                        title=epc.citation.title,
                        uuid=str(epc.uuid),
                    ),
                )
            ),
        ),
        faces_per_cell=ro.ResqmlJaggedArray(
            elements=ro.IntegerHdf5Array(
                null_value=-1,
                values=ro.Hdf5Dataset(
                    path_in_hdf_file=f"/RESQML/{str(hexa_uuid)}/faces_per_cell",
                    hdf_proxy=ro.DataObjectReference(
                        content_type=f"application/x-eml+xml;version={schema_version};type={get_data_object_type(epc)}",
                        title=epc.citation.title,
                        uuid=str(epc.uuid),
                    ),
                )
            ),
            cumulative_length=ro.IntegerHdf5Array(
                null_value=-1,
                values=ro.Hdf5Dataset(
                    path_in_hdf_file=f"/RESQML/{str(hexa_uuid)}/faces_per_cell_cl",
                    hdf_proxy=ro.DataObjectReference(
                        content_type=f"application/x-eml+xml;version={schema_version};type={get_data_object_type(epc)}",
                        title=epc.citation.title,
                        uuid=str(epc.uuid),
                    ),
                )
            ),
        ),
        cell_face_is_right_handed=ro.BooleanHdf5Array(
            values=ro.Hdf5Dataset(
                path_in_hdf_file=f"/RESQML/{str(hexa_uuid)}/cell_face_is_right_handed",
                hdf_proxy=ro.DataObjectReference(
                    content_type=f"application/x-eml+xml;version={schema_version};type={get_data_object_type(epc)}",
                    title=epc.citation.title,
                    uuid=str(epc.uuid),
                ),
            )
        )
    )

    #
    uns = ro.UnstructuredGridRepresentation(
        uuid=str(hexa.uuid),
        schema_version=schema_version,
        # surface_role=resqml_objects.SurfaceRole.MAP,
        citation=create_common_citation(hexa.title),
        cell_count=hexa.cell_count or -1,
        geometry=geom,
    )

    return uns, crs, epc, timeseries, hexa


def convert_epc_mesh_property_to_resqml_mesh(epc_filename, hexa, prop_title, uns, epc, timeseries=None, time_indices:list[int]=[]):
    import resqpy.model as rq
    import resqpy.property as rqp

    def uom_for_prop_title(pt: str):
        if (pt == "Age"):
            return ro.ResqmlUom.A_1
        if (pt == "Temperature"):
            return ro.ResqmlUom.DEG_C
        if (pt == "LayerID"):
            return ro.ResqmlUom.EUC
        if (pt == "Porosity_initial"):
            return ro.ResqmlUom.M3_M3
        if (pt == "Porosity_decay"):
            return ro.ResqmlUom.VALUE_1_M
        if (pt == "Density_solid"):
            return ro.ResqmlUom.KG_M3
        if (pt == "insulance_thermal"):
            return ro.ThermalInsulanceUom.DELTA_K_M2_W
        if (pt == "Radiogenic_heat_production"):
            return ro.ResqmlUom.U_W_M3
        if (pt == 'dynamic nodes') or (pt=='points'):
            return ro.ResqmlUom.M
        if (pt == 'thermal_conductivity'):
            return ro.ResqmlUom.W_M_K
        if (pt == 'Vitrinite reflectance' or pt == '%Ro'):
            return ro.ResqmlUom.VALUE
        if ("Expelled" in pt):
            return ro.ResqmlUom.KG_M3
        if ("Transformation" in pt):
            return ro.ResqmlUom.VALUE
        return ro.ResqmlUom.EUC

    model = rq.Model(epc_filename)
    assert model is not None
    prop_types = ['obj_ContinuousProperty', 'obj_DiscreteProperty', 'obj_CategoricalProperty', 'obj_PointsProperty']
    p = []
    for i in prop_types:
        p1 = model.uuids(title=prop_title, obj_type=i)
        p.extend(p1)
    p_test = rqp.Property(model, uuid=p[0])

    use_timeseries = isinstance(p_test.time_index(), int)
    if not use_timeseries:
        prop_uuid0 = p[0]
        prop0 = rqp.Property(model, uuid=prop_uuid0)
    else:
        prop_uuids = p
        prop_uuid0 = prop_uuids[time_indices[0]]
        prop0 = rqp.Property(model, uuid=prop_uuid0)   # a prop representative of all in the timeseries

    continuous = prop0.is_continuous()

    if (prop0.local_property_kind_uuid() is None):
        propertykind0 = None
    else:
        pk = rqp.PropertyKind(model, uuid=prop0.local_property_kind_uuid())
        propertykind0 = ro.PropertyKind(
            schema_version=schema_version,
            citation=create_common_citation(f"{prop_title}"),
            naming_system="urn:resqml:bp.com:resqpy",
            is_abstract=False,
            representative_uom=uom_for_prop_title(prop_title),
            parent_property_kind=ro.StandardPropertyKind(
                kind=ro.ResqmlPropertyKind.CONTINUOUS if continuous else ro.ResqmlPropertyKind.DISCRETE
            ),
            uuid=str(pk.uuid),
        )


    cprop0s, props = [], []

    for i in range( len(time_indices) if use_timeseries else 1 ):
        if (not use_timeseries):
            prop_uuid = prop_uuid0
            prop = prop0
        else:
            prop_uuid = prop_uuids[time_indices[i]]
            prop = rqp.Property(model, uuid=prop_uuid)

        pov = ro.PatchOfValues(
            values=ro.DoubleHdf5Array(
                values=ro.Hdf5Dataset(
                    path_in_hdf_file=f"/RESQML/{str(prop_uuid)}/values",
                    hdf_proxy=ro.DataObjectReference(
                        content_type=f"application/x-eml+xml;version={schema_version};type={get_data_object_type(epc)}",
                        title=epc.citation.title,
                        uuid=str(epc.uuid),
                    ),
                )
            ) if continuous else
            ro.IntegerHdf5Array(
                values=ro.Hdf5Dataset(
                    path_in_hdf_file=f"/RESQML/{str(prop_uuid)}/values",
                    hdf_proxy=ro.DataObjectReference(
                        content_type=f"application/x-eml+xml;version={schema_version};type={get_data_object_type(epc)}",
                        title=epc.citation.title,
                        uuid=str(epc.uuid),
                    ),
                ),
                null_value=int(1e30),
            )
        )

        timeindex_ref = None
        if use_timeseries:
            time_index = time_indices[i]
            timeindex_ref = ro.TimeIndex(
                index = time_index,
                time_series = ro.DataObjectReference(
                    content_type=f"application/x-resqml+xml;version={schema_version};type={get_data_object_type(timeseries)}",
                    title=timeseries.citation.title,
                    uuid=timeseries.uuid,
                )
            )

        r_uom = ro.ResqmlUom( value= uom_for_prop_title(prop_title) ) if (prop.uom() is None) else prop.uom()

        if (continuous):
            cprop0 = ro.ContinuousProperty(
                schema_version=schema_version,
                citation=create_common_citation(f"{prop_title}"),
                uuid=str(prop.uuid),
                uom = r_uom,
                count=1,
                indexable_element=prop.indexable_element(),
                supporting_representation=ro.DataObjectReference(
                    content_type=f"application/x-resqml+xml;version={schema_version};type={get_data_object_type(uns)}",
                    title=uns.citation.title,
                    uuid=uns.uuid,
                ),
                property_kind=ro.LocalPropertyKind(
                    local_property_kind=ro.DataObjectReference(
                        content_type=f"application/x-resqml+xml;version={schema_version};type={get_data_object_type(propertykind0)}",
                        title=propertykind0.citation.title,
                        uuid=propertykind0.uuid,
                    )
                ) if (propertykind0 is not None) else ro.StandardPropertyKind(kind=prop.property_kind()),
                minimum_value=[prop.minimum_value() or 0.0],
                maximum_value=[prop.maximum_value() or 1.0],
                facet=[ro.PropertyKindFacet(
                    facet=ro.Facet.WHAT,
                    value=prop_title,  # prop.facet(),
                )],
                patch_of_values=[pov],
                time_index=timeindex_ref,
            )
        else:
            cprop0 = ro.DiscreteProperty(
                schema_version=schema_version,
                citation=create_common_citation(f"{prop_title}"),
                uuid=str(prop.uuid),
                # uom = prop.uom(),
                count=1,
                indexable_element=prop.indexable_element(),
                supporting_representation=ro.DataObjectReference(
                    content_type=f"application/x-resqml+xml;version={schema_version};type={get_data_object_type(uns)}",
                    title=uns.citation.title,
                    uuid=uns.uuid,
                ),
                property_kind=ro.LocalPropertyKind(
                    local_property_kind=ro.DataObjectReference(
                        content_type=f"application/x-resqml+xml;version={schema_version};type={get_data_object_type(propertykind0)}",
                        title=propertykind0.citation.title,
                        uuid=propertykind0.uuid,
                    )
                ) if (propertykind0 is not None) else ro.StandardPropertyKind(kind=prop.property_kind()),
                minimum_value=[int(prop.minimum_value() or 0)],
                maximum_value=[int(prop.maximum_value() or 1)],
                facet=[ro.PropertyKindFacet(
                    facet=ro.Facet.WHAT,
                    value=prop_title,  # prop.facet(),
                )],
                patch_of_values=[pov],
                time_index=timeindex_ref,
            )
        cprop0s.append(cprop0)
        props.append(prop)

    return cprop0s, props, propertykind0